# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import sys
from functools import reduce
from itertools import zip_longest

import numpy as np
import pytest

from hypothesis import (
    HealthCheck,
    Phase,
    assume,
    given,
    note,
    settings,
    strategies as st,
    target,
)
from hypothesis.errors import InvalidArgument, UnsatisfiedAssumption
from hypothesis.extra import numpy as nps
from hypothesis.strategies._internal.lazy import unwrap_strategies

from tests.common.debug import (
    assert_all_examples,
    check_can_generate_examples,
    find_any,
    minimal,
)
from tests.common.utils import fails_with, flaky

ANY_SHAPE = nps.array_shapes(min_dims=0, max_dims=32, min_side=0, max_side=32)
ANY_NONZERO_SHAPE = nps.array_shapes(min_dims=0, max_dims=32, min_side=1, max_side=32)


@given(nps.arrays(float, ()))
def test_empty_dimensions_are_arrays(x):
    assert isinstance(x, np.ndarray)
    assert x.dtype.kind == "f"


@given(nps.arrays(float, (1, 0, 1)))
def test_can_handle_zero_dimensions(x):
    assert x.shape == (1, 0, 1)


@given(nps.arrays("uint32", (5, 5)))
def test_generates_unsigned_ints(x):
    assert (x >= 0).all()


@given(nps.arrays(int, (1,)))
def test_assert_fits_in_machine_size(x):
    pass


def test_generates_and_minimizes():
    assert (minimal(nps.arrays(float, (2, 2))) == np.zeros(shape=(2, 2))).all()


def test_can_minimize_large_arrays():
    x = minimal(nps.arrays("uint32", 100), lambda x: np.any(x) and not np.all(x))
    assert np.logical_or(x == 0, x == 1).all()
    assert np.count_nonzero(x) in (1, len(x) - 1)


@flaky(max_runs=50, min_passes=1)
def test_can_minimize_float_arrays():
    with np.errstate(over="ignore", invalid="ignore"):
        x = minimal(nps.arrays(float, 50), lambda t: np.nansum(t) >= 1.0)
        assert x.sum() in (1, 50)


class Foo:
    pass


foos = st.tuples().map(lambda _: Foo())


def test_can_create_arrays_of_composite_types():
    arr = minimal(nps.arrays(object, 100, elements=foos))
    for x in arr:
        assert isinstance(x, Foo)


@given(st.lists(st.integers()), st.data())
def test_can_create_zero_dim_arrays_of_lists(x, data):
    arr = data.draw(nps.arrays(object, (), elements=st.just(x)))
    assert arr.shape == ()
    assert arr.dtype == np.dtype(object)
    assert arr.item() == x


def test_can_create_arrays_of_tuples():
    arr = minimal(
        nps.arrays(object, 10, elements=st.tuples(st.integers(), st.integers())),
        lambda x: all(t0 != t1 for t0, t1 in x),
    )
    assert all(a in ((1, 0), (0, 1)) for a in arr)


@given(nps.arrays(object, (2, 2), elements=st.tuples(st.integers())))
def test_does_not_flatten_arrays_of_tuples(arr):
    assert isinstance(arr[0][0], tuple)


@given(
    nps.arrays(object, (2, 2), elements=st.lists(st.integers(), min_size=1, max_size=1))
)
def test_does_not_flatten_arrays_of_lists(arr):
    assert isinstance(arr[0][0], list)


@given(nps.array_shapes())
def test_can_generate_array_shapes(shape):
    assert isinstance(shape, tuple)
    assert all(isinstance(i, int) for i in shape)


@settings(
    deadline=None, max_examples=10, suppress_health_check=[HealthCheck.nested_given]
)
@given(st.integers(0, 10), st.integers(0, 9), st.integers(0), st.integers(0))
def test_minimise_array_shapes(min_dims, dim_range, min_side, side_range):
    smallest = minimal(
        nps.array_shapes(
            min_dims=min_dims,
            max_dims=min_dims + dim_range,
            min_side=min_side,
            max_side=min_side + side_range,
        )
    )
    assert len(smallest) == min_dims
    assert all(k == min_side for k in smallest)


@pytest.mark.parametrize(
    "kwargs", [{"min_side": 100}, {"min_dims": 15}, {"min_dims": 32}]
)
def test_interesting_array_shapes_argument(kwargs):
    check_can_generate_examples(nps.array_shapes(**kwargs))


@given(nps.scalar_dtypes())
def test_can_generate_scalar_dtypes(dtype):
    assert isinstance(dtype, np.dtype)


@settings(max_examples=100)
@given(
    nps.nested_dtypes(
        subtype_strategy=st.one_of(
            nps.scalar_dtypes(), nps.byte_string_dtypes(), nps.unicode_string_dtypes()
        )
    )
)
def test_can_generate_compound_dtypes(dtype):
    assert isinstance(dtype, np.dtype)


@settings(max_examples=100)
@given(
    nps.nested_dtypes(
        subtype_strategy=st.one_of(
            nps.scalar_dtypes(), nps.byte_string_dtypes(), nps.unicode_string_dtypes()
        )
    ).flatmap(lambda dt: nps.arrays(dtype=dt, shape=1))
)
def test_can_generate_data_compound_dtypes(arr):
    # This is meant to catch the class of errors which prompted PR #2085
    assert isinstance(arr, np.ndarray)


@given(nps.nested_dtypes())
def test_np_dtype_is_idempotent(dtype):
    assert dtype == np.dtype(dtype)


def test_minimise_scalar_dtypes():
    assert minimal(nps.scalar_dtypes()) == np.dtype("bool")


def test_minimise_nested_types():
    assert minimal(nps.nested_dtypes()) == np.dtype("bool")


def test_minimise_array_strategy():
    smallest = minimal(
        nps.arrays(
            nps.nested_dtypes(max_itemsize=200),
            nps.array_shapes(max_dims=3, max_side=3),
        )
    )
    assert smallest.dtype == np.dtype("bool")
    assert not smallest.any()


@given(nps.array_dtypes(allow_subarrays=False))
def test_can_turn_off_subarrays(dt):
    for name in dt.names:
        assert dt.fields[name][0].shape == ()


def test_array_dtypes_may_have_field_titles():
    find_any(nps.array_dtypes(), lambda dt: len(dt.fields) > len(dt.names))


@pytest.mark.parametrize("byteorder", ["<", ">"])
@given(data=st.data())
def test_can_restrict_endianness(data, byteorder):
    dtype = data.draw(nps.integer_dtypes(endianness=byteorder, sizes=(16, 32, 64)))
    if byteorder == ("<" if sys.byteorder == "little" else ">"):
        assert dtype.byteorder == "="
    else:
        assert dtype.byteorder == byteorder


@given(nps.integer_dtypes(sizes=8))
def test_can_specify_size_as_an_int(dt):
    assert dt.itemsize == 1


@given(st.data())
def test_can_draw_arrays_from_scalars(data):
    dt = data.draw(nps.scalar_dtypes())
    result = data.draw(nps.arrays(dtype=dt, shape=()))
    assert isinstance(result, np.ndarray)
    assert result.dtype == dt


@given(st.data())
def test_can_cast_for_arrays(data):
    # Note: this only passes with castable datatypes, certain dtype
    # combinations will result in an error if numpy is not able to cast them.
    dt_elements = np.dtype(data.draw(st.sampled_from(["bool", "<i2", ">i2"])))
    dt_desired = np.dtype(
        data.draw(st.sampled_from(["<i2", ">i2", "float32", "float64"]))
    )
    result = data.draw(
        nps.arrays(
            dtype=dt_desired, elements=nps.from_dtype(dt_elements), shape=(1, 2, 3)
        )
    )
    assert isinstance(result, np.ndarray)
    assert result.dtype == dt_desired


@given(nps.arrays(dtype="int8", shape=st.integers(0, 20), unique=True))
def test_array_values_are_unique(arr):
    assert len(set(arr)) == len(arr)


def test_cannot_generate_unique_array_of_too_many_elements():
    strat = nps.arrays(dtype=int, elements=st.integers(0, 5), shape=10, unique=True)
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(strat)


@given(
    nps.arrays(
        elements=st.just(0.0),
        dtype=float,
        fill=st.just(np.nan),
        shape=st.integers(0, 20),
        unique=True,
    )
)
def test_array_values_are_unique_high_collision(arr):
    assert (arr == 0.0).sum() <= 1


@given(nps.arrays(dtype="int8", shape=(4,), elements=st.integers(0, 3), unique=True))
def test_generates_all_values_for_unique_array(arr):
    # Ensures that the "reject already-seen element" branch is covered
    assert len(set(arr)) == len(arr)


@given(nps.arrays(dtype="int8", shape=255, unique=True))
def test_efficiently_generates_all_unique_array(arr):
    # Avoids the birthday paradox with UniqueSampledListStrategy
    assert len(set(arr)) == len(arr)


@given(st.data(), st.integers(-100, 100), st.integers(1, 100))
def test_array_element_rewriting(data, start, size):
    arr = nps.arrays(
        dtype=np.dtype("int64"),
        shape=size,
        elements=st.integers(start, start + size - 1),
        unique=True,
    )
    assert set(data.draw(arr)) == set(range(start, start + size))


def test_may_fill_with_nan_when_unique_is_set():
    find_any(
        nps.arrays(
            dtype=float,
            elements=st.floats(allow_nan=False),
            shape=10,
            unique=True,
            fill=st.just(np.nan),
        ),
        lambda x: np.isnan(x).any(),
    )


@given(
    nps.arrays(
        dtype=float,
        elements=st.floats(allow_nan=False),
        shape=10,
        unique=True,
        fill=st.just(np.nan),
    )
)
def test_is_still_unique_with_nan_fill(xs):
    assert len(set(xs)) == len(xs)


@fails_with(InvalidArgument)
@given(
    nps.arrays(
        dtype=float,
        elements=st.floats(allow_nan=False),
        shape=10,
        unique=True,
        fill=st.just(0.0),
    )
)
def test_may_not_fill_with_non_nan_when_unique_is_set(arr):
    pass


@fails_with(InvalidArgument)
@given(nps.arrays(dtype="U", shape=10, unique=True, fill=st.just("")))
def test_may_not_fill_with_non_nan_when_unique_is_set_and_type_is_not_number(arr):
    pass


np_version = tuple(int(x) for x in np.__version__.split(".")[:2])


@pytest.mark.parametrize("fill", [False, True])
# Overflowing elements deprecated upstream in Numpy 1.24 :-)
@fails_with(
    InvalidArgument
    if np_version < (1, 24)
    else (DeprecationWarning if np_version < (2, 0) else OverflowError)
)
@given(st.data())
def test_overflowing_integers_are_deprecated(fill, data):
    kw = {"elements": st.just(300)}
    if fill:
        kw = {"elements": st.nothing(), "fill": kw["elements"]}
    arr = data.draw(nps.arrays(dtype="int8", shape=(1,), **kw))
    assert arr[0] == (300 % 256)


@pytest.mark.parametrize("fill", [False, True])
@pytest.mark.parametrize(
    "dtype,strat",
    [
        ("float16", st.floats(min_value=65520, allow_infinity=False)),
        ("float32", st.floats(min_value=10**40, allow_infinity=False)),
        (
            "complex64",
            st.complex_numbers(min_magnitude=10**300, allow_infinity=False),
        ),
        ("U1", st.text(min_size=2, max_size=2)),
        ("S1", st.binary(min_size=2, max_size=2)),
    ],
)
@fails_with(InvalidArgument)
@given(data=st.data())
def test_unrepresentable_elements_are_deprecated(fill, dtype, strat, data):
    if fill:
        kw = {"elements": st.nothing(), "fill": strat}
    else:
        kw = {"elements": strat}
    try:
        arr = data.draw(nps.arrays(dtype=dtype, shape=(1,), **kw))
    except RuntimeWarning:
        assert np_version >= (1, 24), "New overflow-on-cast detection"
        raise InvalidArgument("so the test passes") from None

    try:
        # This is a float or complex number, and has overflowed to infinity,
        # triggering our deprecation for overflow.
        assert np.isinf(arr[0])
    except TypeError:
        # We tried to call isinf on a string.  The string was generated at
        # length two, then truncated by the dtype of size 1 - deprecation
        # again.  If the first character was \0 it is now the empty string.
        assert len(arr[0]) <= 1


@given(nps.arrays(dtype="float16", shape=(1,)))
def test_inferred_floats_do_not_overflow(arr):
    pass


@given(nps.arrays(dtype="float16", shape=10, elements={"min_value": 0, "max_value": 1}))
def test_inferred_floats_can_be_constrained_at_low_width(arr):
    assert (arr >= 0).all()
    assert (arr <= 1).all()


@given(
    nps.arrays(
        dtype="float16",
        shape=10,
        elements={
            "min_value": 0,
            "max_value": 1,
            "exclude_min": True,
            "exclude_max": True,
        },
    )
)
def test_inferred_floats_can_be_constrained_at_low_width_excluding_endpoints(arr):
    assert (arr > 0).all()
    assert (arr < 1).all()


@given(
    nps.arrays(
        dtype="float16",
        shape=10,
        unique=True,
        elements=st.integers(1, 9),
        fill=st.just(np.nan),
    )
)
def test_unique_array_with_fill_can_use_all_elements(arr):
    assume(len(set(arr)) == arr.size)


@given(nps.arrays(dtype="uint8", shape=25, unique=True, fill=st.nothing()))
def test_unique_array_without_fill(arr):
    # This test covers the collision-related branches for fully dense unique arrays.
    # Choosing 25 of 256 possible elements means we're almost certain to see collisions
    # thanks to the 'birthday paradox', but finding unique elemennts is still easy.
    assume(len(set(arr)) == arr.size)


@given(ndim=st.integers(0, 5), data=st.data())
def test_mapped_positive_axes_are_unique(ndim, data):
    min_size = data.draw(st.integers(0, ndim), label="min_size")
    max_size = data.draw(st.integers(min_size, ndim), label="max_size")
    axes = data.draw(
        nps.valid_tuple_axes(ndim, min_size=min_size, max_size=max_size), label="axes"
    )
    assert len(set(axes)) == len({i if 0 < i else ndim + i for i in axes})


@given(ndim=st.integers(0, 5), data=st.data())
def test_length_bounds_are_satisfied(ndim, data):
    min_size = data.draw(st.integers(0, ndim), label="min_size")
    max_size = data.draw(st.integers(min_size, ndim), label="max_size")
    axes = data.draw(
        nps.valid_tuple_axes(ndim, min_size=min_size, max_size=max_size), label="axes"
    )
    assert min_size <= len(axes) <= max_size


@given(shape=nps.array_shapes(), data=st.data())
def test_axes_are_valid_inputs_to_sum(shape, data):
    x = np.zeros(shape, dtype="uint8")
    axes = data.draw(nps.valid_tuple_axes(ndim=len(shape)), label="axes")
    np.sum(x, axes)


@settings(
    deadline=None, max_examples=10, suppress_health_check=[HealthCheck.nested_given]
)
@given(ndim=st.integers(0, 3), data=st.data())
def test_minimize_tuple_axes(ndim, data):
    min_size = data.draw(st.integers(0, ndim), label="min_size")
    max_size = data.draw(st.integers(min_size, ndim), label="max_size")
    smallest = minimal(nps.valid_tuple_axes(ndim, min_size=min_size, max_size=max_size))
    assert len(smallest) == min_size
    assert all(k > -1 for k in smallest)


@settings(
    deadline=None, max_examples=10, suppress_health_check=[HealthCheck.nested_given]
)
@given(ndim=st.integers(0, 3), data=st.data())
def test_minimize_negative_tuple_axes(ndim, data):
    min_size = data.draw(st.integers(0, ndim), label="min_size")
    max_size = data.draw(st.integers(min_size, ndim), label="max_size")
    smallest = minimal(
        nps.valid_tuple_axes(ndim, min_size=min_size, max_size=max_size),
        lambda x: all(i < 0 for i in x),
    )
    assert len(smallest) == min_size


@given(nps.broadcastable_shapes((), min_side=0, max_side=0, min_dims=0, max_dims=0))
def test_broadcastable_empty_shape(shape):
    assert shape == ()


@settings(deadline=None, suppress_health_check=[HealthCheck.too_slow])
@given(shape=ANY_SHAPE, data=st.data())
def test_broadcastable_shape_bounds_are_satisfied(shape, data):
    min_dims = data.draw(st.integers(0, 32), label="min_dims")
    max_dims = data.draw(st.none() | st.integers(min_dims, 32), label="max_dims")
    min_side = data.draw(st.integers(0, 3), label="min_side")
    max_side = data.draw(st.none() | st.integers(min_side, 6), label="max_side")
    try:
        bshape = data.draw(
            nps.broadcastable_shapes(
                shape,
                min_side=min_side,
                max_side=max_side,
                min_dims=min_dims,
                max_dims=max_dims,
            ),
            label="bshape",
        )
    except InvalidArgument:
        raise UnsatisfiedAssumption from None

    if max_dims is None:
        max_dims = max(len(shape), min_dims) + 2

    if max_side is None:
        max_side = max((*shape[::-1][:max_dims], min_side)) + 2

    assert isinstance(bshape, tuple)
    assert all(isinstance(s, int) for s in bshape)
    assert min_dims <= len(bshape) <= max_dims
    assert all(min_side <= s <= max_side for s in bshape)


@settings(deadline=None)
@given(num_shapes=st.integers(1, 4), base_shape=ANY_SHAPE, data=st.data())
def test_mutually_broadcastable_shape_bounds_are_satisfied(
    num_shapes, base_shape, data
):
    min_dims = data.draw(st.integers(0, 32), label="min_dims")
    max_dims = data.draw(
        st.one_of(st.none(), st.integers(min_dims, 32)), label="max_dims"
    )
    min_side = data.draw(st.integers(0, 3), label="min_side")
    max_side = data.draw(
        st.one_of(st.none(), st.integers(min_side, 6)), label="max_side"
    )
    try:
        shapes, result = data.draw(
            nps.mutually_broadcastable_shapes(
                num_shapes=num_shapes,
                base_shape=base_shape,
                min_side=min_side,
                max_side=max_side,
                min_dims=min_dims,
                max_dims=max_dims,
            ),
            label="shapes, result",
        )
    except InvalidArgument:
        raise UnsatisfiedAssumption from None

    if max_dims is None:
        max_dims = max(len(base_shape), min_dims) + 2

    if max_side is None:
        max_side = max((*base_shape[::-1][:max_dims], min_side)) + 2

    assert isinstance(shapes, tuple)
    assert isinstance(result, tuple)
    assert all(isinstance(s, int) for s in result)

    for bshape in shapes:
        assert isinstance(bshape, tuple)
        assert all(isinstance(s, int) for s in bshape)
        assert min_dims <= len(bshape) <= max_dims
        assert all(min_side <= s <= max_side for s in bshape)


def _draw_valid_bounds(data, shape, max_dims, *, permit_none=True):
    if max_dims == 0 or not shape:
        return 0, None

    smallest_side = min(shape[::-1][:max_dims])
    min_strat = (
        st.sampled_from([1, smallest_side])
        if smallest_side > 1
        else st.just(smallest_side)
    )
    min_side = data.draw(min_strat, label="min_side")
    largest_side = max(max(shape[::-1][:max_dims]), min_side)
    if permit_none:
        max_strat = st.one_of(st.none(), st.integers(largest_side, largest_side + 2))
    else:
        max_strat = st.integers(largest_side, largest_side + 2)
    max_side = data.draw(max_strat, label="max_side")
    return min_side, max_side


def _broadcast_two_shapes(shape_a: nps.Shape, shape_b: nps.Shape) -> nps.Shape:
    result = []
    for a, b in zip_longest(reversed(shape_a), reversed(shape_b), fillvalue=1):
        if a != b and (a != 1) and (b != 1):
            raise ValueError(
                f"shapes {shape_a!r} and {shape_b!r} are not broadcast-compatible"
            )
        result.append(a if a != 1 else b)
    return tuple(reversed(result))


def _broadcast_shapes(*shapes):
    """Returns the shape resulting from broadcasting the
    input shapes together.

    Raises ValueError if the shapes are not broadcast-compatible"""
    assert shapes, "Must pass >=1 shapes to broadcast"
    return reduce(_broadcast_two_shapes, shapes, ())


@settings(deadline=None, max_examples=500)
@given(
    shapes=st.lists(
        nps.array_shapes(min_dims=0, min_side=0, max_dims=4, max_side=4), min_size=1
    )
)
def test_broadcastable_shape_util(shapes):
    """Ensures that `_broadcast_shapes` raises when fed incompatible shapes,
    and ensures that it produces the true broadcasted shape"""
    if len(shapes) == 1:
        assert _broadcast_shapes(*shapes) == shapes[0]
        return

    arrs = [np.zeros(s, dtype=np.uint8) for s in shapes]

    try:
        broadcast_out = np.broadcast_arrays(*arrs)
    except ValueError:
        with pytest.raises(ValueError):
            _broadcast_shapes(*shapes)
        return
    broadcasted_shape = _broadcast_shapes(*shapes)

    assert broadcast_out[0].shape == broadcasted_shape


@settings(deadline=None, max_examples=200)
@given(shape=ANY_NONZERO_SHAPE, data=st.data())
def test_broadcastable_shape_has_good_default_values(shape, data):
    # This test ensures that default parameters can always produce broadcast-compatible shapes
    broadcastable_shape = data.draw(
        nps.broadcastable_shapes(shape), label="broadcastable_shapes"
    )
    # error if drawn shape for b is not broadcast-compatible
    _broadcast_shapes(shape, broadcastable_shape)


@settings(deadline=None, max_examples=200)
@given(base_shape=ANY_SHAPE, num_shapes=st.integers(1, 10), data=st.data())
def test_mutually_broadcastableshapes_has_good_default_values(
    num_shapes, base_shape, data
):
    # This test ensures that default parameters can always produce broadcast-compatible shapes
    shapes, result = data.draw(
        nps.mutually_broadcastable_shapes(num_shapes=num_shapes, base_shape=base_shape),
        label="shapes, result",
    )
    assert len(shapes) == num_shapes
    # raises if shapes are not mutually-compatible
    assert result == _broadcast_shapes(base_shape, *shapes)


@settings(deadline=None)
@given(min_dims=st.integers(0, 32), shape=ANY_SHAPE, data=st.data())
def test_broadcastable_shape_can_broadcast(min_dims, shape, data):
    max_dims = data.draw(st.none() | st.integers(min_dims, 32), label="max_dims")
    min_side, max_side = _draw_valid_bounds(data, shape, max_dims)
    broadcastable_shape = data.draw(
        nps.broadcastable_shapes(
            shape,
            min_side=min_side,
            max_side=max_side,
            min_dims=min_dims,
            max_dims=max_dims,
        ),
        label="broadcastable_shapes",
    )
    # error if drawn shape for b is not broadcast-compatible
    _broadcast_shapes(shape, broadcastable_shape)


@settings(deadline=None)
@given(
    num_shapes=st.integers(1, 10),
    min_dims=st.integers(0, 32),
    base_shape=ANY_SHAPE,
    data=st.data(),
)
def test_mutually_broadcastable_shape_can_broadcast(
    num_shapes, min_dims, base_shape, data
):
    max_dims = data.draw(st.none() | st.integers(min_dims, 32), label="max_dims")
    min_side, max_side = _draw_valid_bounds(data, base_shape, max_dims)
    shapes, result = data.draw(
        nps.mutually_broadcastable_shapes(
            num_shapes=num_shapes,
            base_shape=base_shape,
            min_side=min_side,
            max_side=max_side,
            min_dims=min_dims,
            max_dims=max_dims,
        ),
        label="shapes, result",
    )

    # error if drawn shapes are not mutually broadcast-compatible
    assert result == _broadcast_shapes(base_shape, *shapes)


@settings(
    deadline=None, max_examples=50, suppress_health_check=[HealthCheck.nested_given]
)
@given(
    num_shapes=st.integers(1, 3),
    min_dims=st.integers(0, 5),
    base_shape=nps.array_shapes(min_dims=0, max_dims=3, min_side=0, max_side=5),
    data=st.data(),
)
def test_minimize_mutually_broadcastable_shape(num_shapes, min_dims, base_shape, data):
    # Ensure aligned dimensions of broadcastable shape minimizes to `(1,) * min_dims`
    max_dims = data.draw(st.none() | st.integers(min_dims, 5), label="max_dims")
    min_side, max_side = _draw_valid_bounds(
        data, base_shape, max_dims, permit_none=False
    )

    if num_shapes > 1:
        # shrinking gets a little bit hairy when we have empty axes
        # and multiple num_shapes
        assume(min_side > 0)

    smallest_shapes, result = minimal(
        nps.mutually_broadcastable_shapes(
            num_shapes=num_shapes,
            base_shape=base_shape,
            min_side=min_side,
            max_side=max_side,
            min_dims=min_dims,
            max_dims=max_dims,
        )
    )
    note(f"smallest_shapes: {smallest_shapes}")
    note(f"result: {result}")
    assert len(smallest_shapes) == num_shapes
    assert result == _broadcast_shapes(base_shape, *smallest_shapes)
    for smallest in smallest_shapes:
        n_leading = max(len(smallest) - len(base_shape), 0)
        n_aligned = max(len(smallest) - n_leading, 0)
        note(f"n_leading: {n_leading}")
        note(f"n_aligned: {n_aligned} {base_shape[-n_aligned:]}")
        expected = [min_side] * n_leading + [
            (min(1, i) if i != 1 else min_side) if min_side <= 1 <= max_side else i
            for i in (base_shape[-n_aligned:] if n_aligned else ())
        ]
        assert tuple(expected) == smallest


@settings(deadline=None)
@given(max_dims=st.integers(4, 6), data=st.data())
def test_broadcastable_shape_adjusts_max_dim_with_explicit_bounds(max_dims, data):
    # Ensures that `broadcastable_shapes` limits itself to satisfiable dimensions
    # Broadcastable values can only be drawn for dims 0-3 for these shapes
    shape = data.draw(st.sampled_from([(5, 3, 2, 1), (0, 3, 2, 1)]), label="shape")
    broadcastable_shape = data.draw(
        nps.broadcastable_shapes(
            shape, min_side=2, max_side=3, min_dims=3, max_dims=max_dims
        ),
        label="broadcastable_shapes",
    )
    assert len(broadcastable_shape) == 3
    # error if drawn shape for b is not broadcast-compatible
    _broadcast_shapes(shape, broadcastable_shape)


@settings(deadline=None)
@given(
    max_side=st.sampled_from([3, None]),
    min_dims=st.integers(0, 4),
    num_shapes=st.integers(1, 3),
    data=st.data(),
)
def test_mutually_broadcastable_shape_adjusts_max_dim_with_default_bounds(
    max_side, min_dims, num_shapes, data
):
    # Ensures that `mutually_broadcastable_shapes` limits itself to satisfiable dimensions
    # when a default `max_dims` is derived.
    base_shape = data.draw(
        st.sampled_from([(5, 3, 2, 1), (0, 3, 2, 1)]), label="base_shape"
    )

    try:
        shapes, result = data.draw(
            nps.mutually_broadcastable_shapes(
                num_shapes=num_shapes,
                base_shape=base_shape,
                min_side=2,
                max_side=max_side,
                min_dims=min_dims,
            ),
            label="shapes, result",
        )
    except InvalidArgument:
        # There is no satisfiable `max_dims` for us to tune
        assert min_dims == 4
        assert max_side == 3 or base_shape[0] == 0
        return

    if max_side == 3 or base_shape[0] == 0:
        assert all(len(s) <= 3 for s in shapes)
    elif min_dims == 4:
        assert all(4 <= len(s) for s in shapes)

    # error if drawn shape for b is not broadcast-compatible
    assert len(shapes) == num_shapes
    assert result == _broadcast_shapes(base_shape, *shapes)


@settings(
    deadline=None, max_examples=10, suppress_health_check=[HealthCheck.nested_given]
)
@given(min_dims=st.integers(0, 32), min_side=st.integers(2, 3), data=st.data())
def test_broadcastable_shape_shrinking_with_singleton_out_of_bounds(
    min_dims, min_side, data
):
    max_dims = data.draw(st.none() | st.integers(min_dims, 32), label="max_dims")
    max_side = data.draw(st.none() | st.integers(min_side, 6), label="max_side")
    shape = data.draw(st.integers(1, 4).map(lambda n: n * (1,)), label="shape")
    smallest = minimal(
        nps.broadcastable_shapes(
            shape,
            min_side=min_side,
            max_side=max_side,
            min_dims=min_dims,
            max_dims=max_dims,
        )
    )
    assert smallest == (min_side,) * min_dims


@settings(
    deadline=None, max_examples=50, suppress_health_check=[HealthCheck.nested_given]
)
@given(
    num_shapes=st.integers(1, 4),
    min_dims=st.integers(0, 4),
    min_side=st.integers(2, 3),
    data=st.data(),
)
def test_mutually_broadcastable_shapes_shrinking_with_singleton_out_of_bounds(
    num_shapes, min_dims, min_side, data
):
    """Ensures that shapes minimize to `(min_side,) * min_dims` when singleton dimensions
    are disallowed."""
    max_dims = data.draw(st.none() | st.integers(min_dims, 4), label="max_dims")
    max_side = data.draw(
        st.one_of(st.none(), st.integers(min_side, 6)), label="max_side"
    )
    ndims = data.draw(st.integers(1, 4), label="ndim")
    base_shape = (1,) * ndims
    smallest_shapes, result = minimal(
        nps.mutually_broadcastable_shapes(
            num_shapes=num_shapes,
            base_shape=base_shape,
            min_side=min_side,
            max_side=max_side,
            min_dims=min_dims,
            max_dims=max_dims,
        )
    )
    note(f"(smallest_shapes, result): {(smallest_shapes, result)}")
    assert len(smallest_shapes) == num_shapes
    assert result == _broadcast_shapes(base_shape, *smallest_shapes)
    for smallest in smallest_shapes:
        assert smallest == (min_side,) * min_dims


@settings(suppress_health_check=[HealthCheck.too_slow])
@given(
    num_shapes=st.integers(1, 4),
    min_dims=st.integers(1, 32),
    max_side=st.integers(1, 6),
    data=st.data(),
)
def test_mutually_broadcastable_shapes_only_singleton_is_valid(
    num_shapes, min_dims, max_side, data
):
    """Ensures that, when all aligned base-shape dim sizes are larger
    than ``max_side``, only singletons can be drawn"""
    max_dims = data.draw(st.integers(min_dims, 32), label="max_dims")
    base_shape = data.draw(
        nps.array_shapes(min_side=max_side + 1, min_dims=1), label="base_shape"
    )
    input_shapes, result = data.draw(
        nps.mutually_broadcastable_shapes(
            num_shapes=num_shapes,
            base_shape=base_shape,
            min_side=1,
            max_side=max_side,
            min_dims=min_dims,
            max_dims=max_dims,
        ),
        label="input_shapes, result",
    )

    assert len(input_shapes) == num_shapes
    assert result == _broadcast_shapes(base_shape, *input_shapes)
    for shape in input_shapes:
        assert all(i == 1 for i in shape[-len(base_shape) :])


@settings(deadline=None, suppress_health_check=[HealthCheck.too_slow])
@given(
    shape=nps.array_shapes(min_dims=0, max_dims=3, min_side=0, max_side=5),
    max_dims=st.integers(0, 6),
    data=st.data(),
)
def test_broadcastable_shape_can_generate_arbitrary_ndims(shape, max_dims, data):
    # ensures that generates shapes can possess any length in [min_dims, max_dims]
    desired_ndim = data.draw(st.integers(0, max_dims), label="desired_ndim")
    min_dims = data.draw(
        st.one_of(st.none(), st.integers(0, desired_ndim)), label="min_dims"
    )
    # check default arg behavior too
    kwargs = {"min_dims": min_dims} if min_dims is not None else {}
    find_any(
        nps.broadcastable_shapes(shape, min_side=0, max_dims=max_dims, **kwargs),
        lambda x: len(x) == desired_ndim,
        settings(max_examples=10**6),
    )


@settings(deadline=None)
@given(
    num_shapes=st.integers(1, 3),
    base_shape=nps.array_shapes(min_dims=0, max_dims=3, min_side=0, max_side=5),
    max_dims=st.integers(0, 4),
    data=st.data(),
)
def test_mutually_broadcastable_shapes_can_generate_arbitrary_ndims(
    num_shapes, base_shape, max_dims, data
):
    # ensures that each generated shape can possess any length in [min_dims, max_dims]
    desired_ndims = data.draw(
        st.lists(st.integers(0, max_dims), min_size=num_shapes, max_size=num_shapes),
        label="desired_ndims",
    )
    min_dims = data.draw(
        st.one_of(st.none(), st.integers(0, min(desired_ndims))), label="min_dims"
    )
    # check default arg behavior too
    kwargs = {"min_dims": min_dims} if min_dims is not None else {}
    find_any(
        nps.mutually_broadcastable_shapes(
            num_shapes=num_shapes,
            base_shape=base_shape,
            min_side=0,
            max_dims=max_dims,
            **kwargs,
        ),
        lambda x: {len(s) for s in x.input_shapes} == set(desired_ndims),
        settings(max_examples=10**6),
    )


@settings(deadline=None, suppress_health_check=list(HealthCheck))
@given(
    base_shape=nps.array_shapes(min_dims=0, max_dims=3, min_side=0, max_side=2),
    max_dims=st.integers(1, 4),
)
def test_mutually_broadcastable_shapes_can_generate_interesting_singletons(
    base_shape, max_dims
):
    find_any(
        nps.mutually_broadcastable_shapes(
            num_shapes=2,
            base_shape=base_shape,
            min_side=0,
            max_dims=max_dims,
        ),
        lambda x: any(a != b for a, b in zip(*(s[::-1] for s in x.input_shapes), strict=False)),  # type: ignore
    )


@pytest.mark.parametrize("base_shape", [(), (0,), (1,), (2,), (1, 2), (2, 1), (2, 2)])
def test_mutually_broadcastable_shapes_can_generate_mirrored_singletons(base_shape):
    def f(shapes: nps.BroadcastableShapes):
        x, y = shapes.input_shapes
        return x.count(1) == 1 and y.count(1) == 1 and x[::-1] == y

    find_any(
        nps.mutually_broadcastable_shapes(
            num_shapes=2,
            base_shape=base_shape,
            min_side=0,
            max_side=3,
            min_dims=2,
            max_dims=2,
        ),
        f,
    )


@settings(deadline=None)
@given(
    shape=nps.array_shapes(min_dims=1, min_side=1),
    dtype=st.one_of(nps.unsigned_integer_dtypes(), nps.integer_dtypes()),
    data=st.data(),
)
def test_advanced_integer_index_is_valid_with_default_result_shape(shape, dtype, data):
    index = data.draw(nps.integer_array_indices(shape, dtype=dtype))
    x = np.zeros(shape)
    out = x[index]  # raises if the index is invalid
    assert not np.shares_memory(x, out)  # advanced indexing should not return a view
    assert all(dtype == x.dtype for x in index)


@settings(deadline=None)
@given(
    shape=nps.array_shapes(min_dims=1, min_side=1),
    min_dims=st.integers(0, 3),
    min_side=st.integers(0, 3),
    dtype=st.one_of(nps.unsigned_integer_dtypes(), nps.integer_dtypes()),
    data=st.data(),
)
def test_advanced_integer_index_is_valid_and_satisfies_bounds(
    shape, min_dims, min_side, dtype, data
):
    max_side = data.draw(st.integers(min_side, min_side + 2), label="max_side")
    max_dims = data.draw(st.integers(min_dims, min_dims + 2), label="max_dims")
    index = data.draw(
        nps.integer_array_indices(
            shape,
            result_shape=nps.array_shapes(
                min_dims=min_dims,
                max_dims=max_dims,
                min_side=min_side,
                max_side=max_side,
            ),
            dtype=dtype,
        )
    )
    x = np.zeros(shape)
    out = x[index]  # raises if the index is invalid
    assert all(min_side <= s <= max_side for s in out.shape)
    assert min_dims <= out.ndim <= max_dims
    assert not np.shares_memory(x, out)  # advanced indexing should not return a view
    assert all(dtype == x.dtype for x in index)


@settings(deadline=None, suppress_health_check=[HealthCheck.nested_given])
@given(
    shape=nps.array_shapes(min_dims=1, min_side=1),
    min_dims=st.integers(0, 3),
    min_side=st.integers(0, 3),
    dtype=st.sampled_from(["uint8", "int8"]),
    data=st.data(),
)
def test_advanced_integer_index_minimizes_as_documented(
    shape, min_dims, min_side, dtype, data
):
    max_side = data.draw(st.integers(min_side, min_side + 2), label="max_side")
    max_dims = data.draw(st.integers(min_dims, min_dims + 2), label="max_dims")
    result_shape = nps.array_shapes(
        min_dims=min_dims, max_dims=max_dims, min_side=min_side, max_side=max_side
    )
    smallest = minimal(
        nps.integer_array_indices(shape, result_shape=result_shape, dtype=dtype)
    )
    desired = len(shape) * (np.zeros(min_dims * [min_side]),)
    assert len(smallest) == len(desired)
    for s, d in zip(smallest, desired, strict=True):
        np.testing.assert_array_equal(s, d)


@settings(
    deadline=None, max_examples=10, suppress_health_check=[HealthCheck.nested_given]
)
@given(
    shape=nps.array_shapes(min_dims=1, max_dims=2, min_side=1, max_side=3),
    data=st.data(),
)
def test_advanced_integer_index_can_generate_any_pattern(shape, data):
    # ensures that generated index-arrays can be used to yield any pattern of elements from an array
    x = np.arange(np.prod(shape)).reshape(shape)

    target_array = data.draw(
        nps.arrays(
            shape=nps.array_shapes(min_dims=1, max_dims=2, min_side=1, max_side=2),
            elements=st.sampled_from(x.flatten()),
            dtype=x.dtype,
        ),
        label="target",
    )

    def index_selects_values_in_order(index):
        selected = x[index]
        target(len(set(selected.flatten())), label="unique indices")
        target(float(np.sum(target_array == selected)), label="elements correct")
        return np.all(target_array == selected)

    minimal(
        nps.integer_array_indices(shape, result_shape=st.just(target_array.shape)),
        index_selects_values_in_order,
        settings(max_examples=10**6, phases=[Phase.generate, Phase.target]),
    )


@pytest.mark.parametrize(
    "condition",
    [
        lambda ix: isinstance(ix, tuple) and Ellipsis in ix,
        lambda ix: isinstance(ix, tuple) and Ellipsis not in ix,
        lambda ix: isinstance(ix, tuple) and np.newaxis in ix,
        lambda ix: isinstance(ix, tuple) and np.newaxis not in ix,
        lambda ix: ix is Ellipsis,
        lambda ix: ix == np.newaxis,
    ],
)
def test_basic_indices_options(condition):
    indexers = nps.array_shapes(min_dims=0, max_dims=32).flatmap(
        lambda shape: nps.basic_indices(shape, allow_newaxis=True)
    )
    find_any(indexers, condition)


def test_basic_indices_can_generate_empty_tuple():
    find_any(nps.basic_indices(shape=(0, 0), allow_ellipsis=True), lambda ix: ix == ())


def test_basic_indices_can_generate_non_tuples():
    find_any(
        nps.basic_indices(shape=(0, 0), allow_ellipsis=True),
        lambda ix: not isinstance(ix, tuple),
    )


def test_basic_indices_can_generate_long_ellipsis():
    # Runs of slice(None) - such as [0,:,:,:,0] - can be replaced by e.g. [0,...,0]
    find_any(
        nps.basic_indices(shape=(1, 0, 0, 0, 1), allow_ellipsis=True),
        lambda ix: len(ix) == 3 and ix[1] == Ellipsis,
    )


@given(
    nps.basic_indices(shape=(0, 0, 0, 0, 0)).filter(
        lambda idx: isinstance(idx, tuple) and Ellipsis in idx
    )
)
def test_basic_indices_replaces_whole_axis_slices_with_ellipsis(idx):
    # `slice(None)` (aka `:`) is the only valid index for an axis of size
    # zero, so if all dimensions are 0 then a `...` will replace all the
    # slices because we generate `...` for entire contiguous runs of `:`
    assert slice(None) not in idx


def test_basic_indices_can_generate_indices_not_covering_all_dims():
    # These "flat indices" are skippable in the underlying BasicIndexStrategy,
    # so we ensure we're definitely generating them for nps.basic_indices().
    find_any(
        nps.basic_indices(shape=(3, 3, 3)),
        lambda ix: (
            (not isinstance(ix, tuple) and ix != Ellipsis)
            or (isinstance(ix, tuple) and Ellipsis not in ix and len(ix) < 3)
        ),
        settings=settings(max_examples=5_000),
    )


@given(
    shape=nps.array_shapes(min_dims=0, max_side=4)
    | nps.array_shapes(min_dims=0, min_side=0, max_side=10),
    allow_newaxis=st.booleans(),
    allow_ellipsis=st.booleans(),
    data=st.data(),
)
def test_basic_indices_generate_valid_indexers(
    shape, allow_newaxis, allow_ellipsis, data
):
    min_dims = data.draw(
        st.integers(0, 5 if allow_newaxis else len(shape)), label="min_dims"
    )
    max_dims = data.draw(
        st.none() | st.integers(min_dims, 32 if allow_newaxis else len(shape)),
        label="max_dims",
    )
    indexer = data.draw(
        nps.basic_indices(
            shape,
            min_dims=min_dims,
            max_dims=max_dims,
            allow_ellipsis=allow_ellipsis,
            allow_newaxis=allow_newaxis,
        ),
        label="indexer",
    )

    # Check that disallowed things are indeed absent
    if not allow_newaxis:
        if isinstance(indexer, tuple):
            assert 0 <= len(indexer) <= len(shape) + int(allow_ellipsis)
        else:
            assert 1 <= len(shape) + int(allow_ellipsis)
        assert np.newaxis not in shape
    if not allow_ellipsis:
        assert Ellipsis not in shape

    if 0 in shape:
        # If there's a zero in the shape, the array will have no elements.
        array = np.zeros(shape)
        assert array.size == 0
    elif np.prod(shape) <= 10**5:
        # If it's small enough to instantiate, do so with distinct elements.
        array = np.arange(np.prod(shape)).reshape(shape)
    else:
        # We can't cheat on this one, so just try another.
        assume(False)
    view = array[indexer]
    if not np.isscalar(view):
        assert min_dims <= view.ndim <= (32 if max_dims is None else max_dims)
        if view.size:
            assert np.shares_memory(view, array)


# addresses https://github.com/HypothesisWorks/hypothesis/issues/2582
@given(
    nps.arrays(
        shape=nps.array_shapes(min_dims=0, min_side=0), dtype=nps.floating_dtypes()
    )
)
def test_array_owns_memory(x: np.ndarray):
    assert x.base is None
    assert x[...].base is x


@given(st.data())
def test_no_recursion_in_multi_line_reprs_issue_3560(data):
    data.draw(nps.arrays(shape=(2,), dtype=float).map(lambda x: x))
    data.draw(
        nps.arrays(
            shape=(2,),
            dtype=float,
        ).map(lambda x: x)
    )


def test_infers_elements_and_fill():
    # Regression test for https://github.com/HypothesisWorks/hypothesis/issues/3900
    # We only infer a fill strategy if the elements_strategy has reusable values,
    # and the interaction of two performance fixes broke this.  Oops...
    s = unwrap_strategies(nps.arrays(dtype=np.uint32, shape=1))
    assert isinstance(s, nps.ArrayStrategy)
    assert repr(s.element_strategy) == f"integers(0, {2**32-1})"
    assert repr(s.fill) == f"integers(0, {2**32-1})"

    # But we _don't_ infer a fill if the elements strategy is non-reusable
    elems = st.builds(lambda x: x * 2, st.integers(1, 10)).map(np.uint32)
    assert not elems.has_reusable_values
    s = unwrap_strategies(nps.arrays(dtype=np.uint32, shape=1, elements=elems))
    assert s.fill.is_empty


@given(nps.arrays(np.dtype("O"), shape=nps.array_shapes()))
def test_object_arrays_are_of_type_object(obj_array):
    assert obj_array.dtype == np.dtype("O")


def test_class_instances_not_allowed_in_scalar_array():
    class A:
        pass

    s = nps.arrays(
        nps.scalar_dtypes(),
        shape=nps.array_shapes(),
        elements=st.just(A()),
    )

    # can raise ValueError during generation. For example if scalar_dtype is
    # corresponds to a datetime, numpy will raise "cannot convert A to a datetime".
    with pytest.raises((InvalidArgument, ValueError)):
        check_can_generate_examples(s)


def test_object_arrays_with_mixed_elements_has_object_dtype():
    class A:
        pass

    s = nps.arrays(
        np.dtype("O"),
        shape=nps.array_shapes(),
        elements=st.just(A()) | st.integers(),
    )

    assert_all_examples(s, lambda arr: arr.dtype == np.dtype("O"))
    find_any(s, lambda arr: len({type(x) for x in arr.ravel()}) > 1)


@given(st.data())
def test_object_array_can_hold_arbitrary_class_instances(data):
    instance = data.draw(st.from_type(type).flatmap(st.from_type))
    s = nps.arrays(np.dtype("O"), nps.array_shapes(), elements=st.just(instance))
    arr = data.draw(s)

    assert all(v is instance for v in arr.ravel())


def test_object_array_can_hold_incomparable_elements():
    class Incomparable:
        def __eq__(self, other):
            raise TypeError

    check_can_generate_examples(
        nps.arrays(
            np.dtype("O"),
            nps.array_shapes(),
            elements=st.just(Incomparable()),
        )
    )


def test_can_generate_nested_object_arrays():
    int_arrays = nps.arrays(np.dtype("int"), nps.array_shapes())
    s = nps.arrays(np.dtype("O"), nps.array_shapes(), elements=int_arrays)
    check_can_generate_examples(s)



#!/usr/bin/env python

# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import hashlib
import sys

if __name__ == "__main__":
    print(hashlib.sha384(sys.stdin.read().encode()).hexdigest()[:10])



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import asyncio
import warnings
from unittest import TestCase

import pytest

from hypothesis import assume, given, settings, strategies as st
from hypothesis.internal.compat import PYPY

from tests.common.utils import skipif_emscripten

pytestmark = pytest.mark.skipif(
    settings.get_current_profile_name() == "threading",
    reason="bad interactions when mixing threads and asyncio",
)


def coro_decorator(f):
    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=DeprecationWarning)
        try:
            return asyncio.coroutine(f)
        except AttributeError:
            pytest.skip("needs fixing for asyncio version", allow_module_level=True)


class TestAsyncio(TestCase):
    timeout = 5

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.loop.close()

    def execute_example(self, f):
        error = None

        def g():
            nonlocal error
            try:
                x = f()
                if x is not None:
                    yield from x
            except BaseException as e:
                error = e

        coro = coro_decorator(g)
        future = asyncio.wait_for(coro(), timeout=self.timeout)
        self.loop.run_until_complete(future)
        if error is not None:
            raise error

    @pytest.mark.skipif(PYPY, reason="Error in asyncio.new_event_loop()")
    @given(x=st.text())
    @coro_decorator
    def test_foo(self, x):
        assume(x)
        yield from asyncio.sleep(0.001)
        assert x


class TestAsyncioRun(TestCase):
    # In principle, these tests could indeed run on emscripten if we grab the existing
    # event loop and run them there.  However, that seems to have hit an infinite loop
    # and so we're just skipping them for now and will revisit later.

    def execute_example(self, f):
        asyncio.run(f())

    @skipif_emscripten
    @given(x=st.text())
    @coro_decorator
    def test_foo_yield_from(self, x):
        assume(x)
        yield from asyncio.sleep(0.001)
        assert x

    @skipif_emscripten
    @given(st.text())
    async def test_foo_await(self, x):
        assume(x)
        await asyncio.sleep(0.001)
        assert x



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

TEST_DECORATORS_ALONE = """
import hypothesis
from hypothesis.strategies import composite, none

@composite
def test_composite_is_not_a_test(draw):
    # This strategy will be instantiated, but no draws == no calls.
    return draw(none())

@hypothesis.seed(0)
def test_seed_without_given_fails():
    pass

@hypothesis.example(x=None)
def test_example_without_given_fails():
    pass

@hypothesis.reproduce_failure(hypothesis.__version__, b"AA==")
def test_repro_without_given_fails():
    pass
"""


def test_decorators_without_given_should_fail(testdir):
    script = testdir.makepyfile(TEST_DECORATORS_ALONE)
    result = testdir.runpytest(script)
    result.assert_outcomes(failed=4)
    assert "pytest_runtest_call" not in "\n".join(result.outlines)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import gc
import inspect
import json
import os
import random
import sys
import time as time_module
from functools import wraps
from pathlib import Path

import pytest
from _pytest.monkeypatch import MonkeyPatch

from hypothesis import is_hypothesis_test, settings
from hypothesis._settings import is_in_ci
from hypothesis.errors import NonInteractiveExampleWarning
from hypothesis.internal import lambda_sources
from hypothesis.internal.compat import add_note
from hypothesis.internal.conjecture import junkdrawer

from tests.common import TIME_INCREMENT
from tests.common.setup import run
from tests.common.utils import raises_warning

run()

# Skip collection of tests which require the Django test runner,
# or that don't work on the current version of Python.
collect_ignore_glob = ["django/*"]
# this checks up until py319
for minor_increment in range(10):
    minor = 10 + minor_increment
    if sys.version_info < (3, minor):
        collect_ignore_glob.append(f"cover/*py3{minor}*.py")

if sys.version_info >= (3, 11):
    collect_ignore_glob.append("cover/test_asyncio.py")  # @asyncio.coroutine removed


in_shrinking_benchmark = False


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: pandas expects this marker to exist.")
    config.addinivalue_line(
        "markers",
        "xp_min_version(api_version): run when greater or equal to api_version",
    )
    config.addinivalue_line("markers", "xf_crosshair: selection for xfailing symbolics")

    if config.getoption("--hypothesis-benchmark-shrinks"):
        # we'd like to support xdist here, but a session-scope fixture won't
        # be enough: https://github.com/pytest-dev/pytest-xdist/issues/271.
        # Need a lockfile or equivalent.

        assert config.getoption(
            "--hypothesis-benchmark-output"
        ), "must specify shrinking output file"

        global in_shrinking_benchmark
        in_shrinking_benchmark = True


def pytest_addoption(parser):
    parser.addoption("--hypothesis-update-outputs", action="store_true")
    parser.addoption("--hypothesis-benchmark-shrinks", type=str, choices=["new", "old"])
    parser.addoption("--hypothesis-benchmark-output", type=str)


@pytest.fixture(scope="function", params=["warns", "raises"])
def warns_or_raises(request):
    """This runs the test twice: first to check that a warning is emitted
    and execution continues successfully despite the warning; then to check
    that the raised warning is handled properly.
    """
    if request.param == "raises":
        return raises_warning
    else:
        return pytest.warns


# crosshair needs actual time for its path timeouts; load it before patching
try:
    import hypothesis_crosshair_provider.crosshair_provider  # noqa: F401
except ImportError:
    pass


if sys.version_info >= (3, 11):
    # To detect if changes in code generation causes lambda test compilation
    # to fail. Older versions (3.10 and earlier) have a few known false
    # negatives which we ignore.
    @pytest.fixture(scope="function", autouse=True)
    def _make_unknown_lambdas_fail(monkeypatch):

        def fail(candidate):
            msg = (
                f"Failed to find a matching source for {candidate}. "
                "This could indicate changes in the Python code generator,\n"
                "or just a previously unknown case. To quickly resolve this "
                "problem, use the `allow_unknown_lambdas` fixture."
            )
            raise AssertionError(msg)

        monkeypatch.setattr(
            lambda_sources, "_check_unknown_perfectly_aligned_lambda", fail
        )


@pytest.fixture(scope="function")
def allow_unknown_lambdas(monkeypatch):
    # Will run after make...fail since autouse are run first

    def nofail(candidate):
        pass

    monkeypatch.setattr(
        lambda_sources, "_check_unknown_perfectly_aligned_lambda", nofail
    )


# monkeypatch is not thread-safe, so pytest-run-parallel will skip all our tests
# if we define this.
if settings.get_current_profile_name() != "threading":

    @pytest.fixture(scope="function", autouse=True)
    def _consistently_increment_time(monkeypatch):
        """Rather than rely on real system time we monkey patch time.time so that
        it passes at a consistent rate between calls.

        The reason for this is that when these tests run in CI, their performance is
        extremely variable and the VM the tests are on might go to sleep for a bit,
        introducing arbitrary delays. This can cause a number of tests to fail
        flakily.

        Replacing time with a fake version under our control avoids this problem.
        """
        frozen = False

        current_time = time_module.time()

        def time():
            nonlocal current_time
            if not frozen:
                current_time += TIME_INCREMENT
            return current_time

        def sleep(naptime):
            nonlocal current_time
            current_time += naptime

        def freeze():
            nonlocal frozen
            frozen = True

        def _patch(name, fn):
            monkeypatch.setattr(
                time_module, name, wraps(getattr(time_module, name))(fn)
            )

        _patch("time", time)
        _patch("monotonic", time)
        _patch("perf_counter", time)
        _patch("sleep", sleep)
        monkeypatch.setattr(time_module, "freeze", freeze, raising=False)

        # In the patched time regime, observing it causes it to increment. To avoid reintroducing
        # non-determinism due to GC running at arbitrary times, we patch the GC observer
        # to NOT increment time.

        monkeypatch.setattr(junkdrawer, "_perf_counter", time)

        if hasattr(gc, "callbacks"):
            # ensure timer callback is added, then bracket it by freeze/unfreeze below
            junkdrawer.gc_cumulative_time()

            _was_frozen = False

            def _freezer(*_):
                nonlocal _was_frozen, frozen
                _was_frozen = frozen
                frozen = True

            def _unfreezer(*_):
                nonlocal _was_frozen, frozen
                frozen = _was_frozen

            gc.callbacks.insert(0, _freezer)  # freeze before gc callback
            gc.callbacks.append(_unfreezer)  # unfreeze after

            yield

            assert gc.callbacks.pop(0) == _freezer
            assert gc.callbacks.pop() == _unfreezer
        else:  # pragma: no cover # branch never taken in CPython
            yield


random_states_after_tests = {}
independent_random = random.Random()


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_call(item):
    if item.config.getoption("--hypothesis-benchmark-shrinks"):
        yield from _benchmark_shrinks(item)
        # ideally benchmark shrinking would not be mutually exclusive with the
        # other checks in this function, but it's cleaner to early-return here,
        # and in practice they will error in normal tests before one runs a
        # benchmark.
        return
    # This hookwrapper checks for PRNG state leaks from Hypothesis tests.
    # See: https://github.com/HypothesisWorks/hypothesis/issues/1919
    if (
        not (hasattr(item, "obj") and is_hypothesis_test(item.obj))
        # we disable this check on the threading job, due to races in the global
        # state.
        or settings.get_current_profile_name() == "threading"
    ):
        outcome = yield
    elif "pytest_randomly" in sys.modules:
        # See https://github.com/HypothesisWorks/hypothesis/issues/3041 - this
        # branch exists to make it easier on external contributors, but should
        # never run in our CI (because that would disable the check entirely).
        assert not is_in_ci()
        outcome = yield
    else:
        # We start by peturbing the state of the PRNG, because repeatedly
        # leaking PRNG state resets state_after to the (previously leaked)
        # state_before, and that just shows as "no use of random".
        random.seed(independent_random.randrange(2**32))
        before = random.getstate()
        outcome = yield
        after = random.getstate()
        if before != after:
            if after in random_states_after_tests:
                raise Exception(
                    f"{item.nodeid!r} and {random_states_after_tests[after]!r} "
                    "both used the `random` module, and finished with the "
                    "same global `random.getstate()`; this is probably a nasty bug!"
                )
            random_states_after_tests[after] = item.nodeid

    # Annotate usage of .example() with a hint about alternatives
    if isinstance(getattr(outcome, "exception", None), NonInteractiveExampleWarning):
        add_note(
            outcome.exception,
            "For hypothesis' own test suite, consider using one of the helper "
            "methods in tests.common.debug instead.",
        )


timer = time_module.process_time


def _worker_path(session: pytest.Session) -> Path:
    return (
        Path(session.config.getoption("--hypothesis-benchmark-output")).parent
        # https://pytest-xdist.readthedocs.io/en/stable/how-to.html#envvar-PYTEST_XDIST_WORKER
        / f"shrinking_results_{os.environ['PYTEST_XDIST_WORKER']}.json"
    )


def _benchmark_shrinks(item: pytest.Function) -> None:
    from hypothesis.internal.conjecture.shrinker import Shrinker

    # this isn't perfect, but it is cheap!
    if "minimal(" not in inspect.getsource(item.function):
        pytest.skip("(probably) does not call minimal()")

    actual_shrink = Shrinker.shrink
    shrink_calls = []
    shrink_time = []

    def shrink(self, *args, **kwargs):
        nonlocal shrink_calls
        nonlocal shrink_time
        start_t = timer()
        result = actual_shrink(self, *args, **kwargs)
        shrink_calls.append(self.engine.call_count - self.initial_calls)
        shrink_time.append(timer() - start_t)
        return result

    monkeypatch = MonkeyPatch()
    monkeypatch.setattr(Shrinker, "shrink", shrink)

    # pytest_runtest_call must yield at some point. This executes the test function
    # n - 1 times, and then yields for the final execution.
    for _ in range(5 - 1):
        item.runtest()
    yield

    monkeypatch.undo()

    # remove leading hypothesis-python/tests/...
    nodeid = item.nodeid.rsplit("/", 1)[1]

    results_p = _worker_path(item.session)
    if not results_p.exists():
        results_p.write_text(json.dumps({"calls": {}, "time": {}}))

    data = json.loads(results_p.read_text())
    data["calls"][nodeid] = shrink_calls
    data["time"][nodeid] = shrink_time
    results_p.write_text(json.dumps(data))


def pytest_sessionfinish(session, exitstatus):
    if not (mode := session.config.getoption("--hypothesis-benchmark-shrinks")):
        return
    # only run on the controller process, not the workers
    if hasattr(session.config, "workerinput"):
        return

    results = {"calls": {}, "time": {}}
    output_p = Path(session.config.getoption("--hypothesis-benchmark-output"))
    for p in output_p.parent.iterdir():
        if p.name.startswith("shrinking_results_"):
            worker_results = json.loads(p.read_text())
            results["calls"] |= worker_results["calls"]
            results["time"] |= worker_results["time"]
            p.unlink()

    results = {mode: results}
    if not output_p.exists():
        output_p.write_text(json.dumps(results))
    else:
        data = json.loads(output_p.read_text())
        data[mode] = results[mode]
        output_p.write_text(json.dumps(data))



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import abc
import enum
import typing
from collections.abc import Callable, Sequence
from inspect import Parameter as P, Signature
from typing import Generic, List as _List, TypeVar, Union

import pytest

from hypothesis import given, infer, settings, strategies as st
from hypothesis.errors import InvalidArgument, ResolutionFailed
from hypothesis.internal.compat import get_type_hints
from hypothesis.internal.reflection import get_pretty_function_description
from hypothesis.strategies._internal import types
from hypothesis.strategies._internal.lazy import LazyStrategy
from hypothesis.strategies._internal.types import _global_type_lookup
from hypothesis.strategies._internal.utils import _all_strategies

from tests.common.debug import (
    assert_all_examples,
    assert_simple_property,
    check_can_generate_examples,
    find_any,
)
from tests.common.utils import (
    Why,
    fails_with,
    skipif_threading,
    temp_registered,
    xfail_on_crosshair,
)

# we'll continue testing the typing variants until their removal from the stdlib
# ruff: noqa: UP006, UP035, UP007

types_with_core_strat = {
    type_
    for type_, strat in _global_type_lookup.items()
    if isinstance(strat, LazyStrategy) and strat.function in vars(st).values()
}


def test_generic_sequence_of_integers_may_be_lists_or_bytes():
    strat = st.from_type(Sequence[int])
    find_any(strat, lambda x: isinstance(x, bytes))
    find_any(strat, lambda x: isinstance(x, list))


@pytest.mark.parametrize("typ", sorted(types_with_core_strat, key=str))
def test_resolve_core_strategies(typ):
    @given(st.from_type(typ))
    def inner(ex):
        assert isinstance(ex, typ)

    inner()


def test_lookup_knows_about_all_core_strategies():
    # Build a set of all types output by core strategies
    blocklist = {
        "builds",
        "data",
        "deferred",
        "from_regex",
        "from_type",
        "ip_addresses",
        "iterables",
        "just",
        "nothing",
        "one_of",
        "permutations",
        "random_module",
        "randoms",
        "recursive",
        "runner",
        "sampled_from",
        "shared",
        "timezone_keys",
        "timezones",
    }
    assert set(_all_strategies).issuperset(blocklist), blocklist.difference(
        _all_strategies
    )
    found = set()
    for thing in (
        getattr(st, name)
        for name in sorted(_all_strategies)
        if name in dir(st) and name not in blocklist
    ):
        for n in range(3):
            try:
                ex = find_any(thing(*([st.nothing()] * n)))
                found.add(type(ex))
                break
            except Exception:
                continue

    cannot_lookup = found - set(types._global_type_lookup)
    assert not cannot_lookup


def test_lookup_keys_are_types():
    with pytest.raises(InvalidArgument):
        st.register_type_strategy("int", st.integers())
    assert "int" not in types._global_type_lookup


@pytest.mark.parametrize(
    "typ, not_a_strategy",
    [
        (int, 42),  # Values must be strategies
        # Can't register NotImplemented directly, even though strategy functions
        # can return it.
        (int, NotImplemented),
    ],
)
def test_lookup_values_are_strategies(typ, not_a_strategy):
    with pytest.raises(InvalidArgument):
        st.register_type_strategy(typ, not_a_strategy)
    assert not_a_strategy not in types._global_type_lookup.values()


@pytest.mark.parametrize("typ", sorted(types_with_core_strat, key=str))
def test_lookup_overrides_defaults(typ):
    sentinel = object()
    with temp_registered(typ, st.just(sentinel)):
        assert_simple_property(st.from_type(typ), lambda v: v is sentinel)
    assert_simple_property(st.from_type(typ), lambda v: v is not sentinel)


class ParentUnknownType:
    pass


class UnknownType(ParentUnknownType):
    def __init__(self, arg):
        pass


def test_custom_type_resolution_fails_without_registering():
    fails = st.from_type(UnknownType)
    with pytest.raises(ResolutionFailed):
        check_can_generate_examples(fails)


def test_custom_type_resolution():
    sentinel = object()
    with temp_registered(UnknownType, st.just(sentinel)):
        assert_simple_property(st.from_type(UnknownType), lambda v: v is sentinel)
        # Also covered by registration of child class
        assert_simple_property(st.from_type(ParentUnknownType), lambda v: v is sentinel)


def test_custom_type_resolution_with_function():
    sentinel = object()
    with temp_registered(UnknownType, lambda _: st.just(sentinel)):
        assert_simple_property(st.from_type(UnknownType), lambda v: v is sentinel)
        assert_simple_property(st.from_type(ParentUnknownType), lambda v: v is sentinel)


def test_custom_type_resolution_with_function_non_strategy():
    with temp_registered(UnknownType, lambda _: None):
        with pytest.raises(ResolutionFailed):
            check_can_generate_examples(st.from_type(UnknownType))
        with pytest.raises(ResolutionFailed):
            check_can_generate_examples(st.from_type(ParentUnknownType))


@pytest.mark.parametrize("strategy_returned", [True, False])
def test_conditional_type_resolution_with_function(strategy_returned):
    sentinel = object()

    def resolve_strategy(thing):
        assert thing == UnknownType
        if strategy_returned:
            return st.just(sentinel)
        return NotImplemented

    with temp_registered(UnknownType, resolve_strategy):
        if strategy_returned:
            assert_simple_property(st.from_type(UnknownType), lambda v: v is sentinel)
        else:
            with pytest.raises(ResolutionFailed):
                check_can_generate_examples(st.from_type(UnknownType))


def test_errors_if_generic_resolves_empty():
    with temp_registered(UnknownType, lambda _: st.nothing()):
        fails_1 = st.from_type(UnknownType)
        with pytest.raises(ResolutionFailed):
            check_can_generate_examples(fails_1)
        fails_2 = st.from_type(ParentUnknownType)
        with pytest.raises(ResolutionFailed):
            check_can_generate_examples(fails_2)


@skipif_threading
def test_cannot_register_empty():
    # Cannot register and did not register
    with pytest.raises(InvalidArgument):
        st.register_type_strategy(UnknownType, st.nothing())
    fails = st.from_type(UnknownType)
    with pytest.raises(ResolutionFailed):
        check_can_generate_examples(fails)
    assert UnknownType not in types._global_type_lookup


def test_pulic_interface_works():
    check_can_generate_examples(st.from_type(int))
    fails = st.from_type("not a type or annotated function")
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(fails)


@pytest.mark.parametrize("infer_token", [infer, ...])
def test_given_can_infer_from_manual_annotations(infer_token):
    # Editing annotations before decorating is hilariously awkward, but works!
    def inner(a):
        assert isinstance(a, int)

    inner.__annotations__ = {"a": int}
    given(a=infer_token)(inner)()


class EmptyEnum(enum.Enum):
    pass


@fails_with(InvalidArgument)
@given(st.from_type(EmptyEnum))
def test_error_if_enum_is_empty(x):
    pass


class BrokenClass:
    __init__ = "Hello!"


def test_uninspectable_builds():
    with pytest.raises(TypeError, match="object is not callable"):
        check_can_generate_examples(st.builds(BrokenClass))


def test_uninspectable_from_type():
    with pytest.raises(TypeError, match="object is not callable"):
        check_can_generate_examples(st.from_type(BrokenClass))


def _check_instances(t):
    # See https://github.com/samuelcolvin/pydantic/discussions/2508
    return (
        t.__module__ != "typing"
        and t.__name__ != "ByteString"
        and not t.__module__.startswith("pydantic")
        and t.__module__ != "typing_extensions"
    )


def maybe_mark(x):
    if x.__name__ in "Match Decimal IPv4Address":
        marks = xfail_on_crosshair(Why.other, as_marks=True, strict=False)
        return pytest.param(x, marks=marks)
    return x


@pytest.mark.parametrize(
    "typ",
    sorted(
        (maybe_mark(x) for x in _global_type_lookup if _check_instances(x)),
        key=str,
    ),
)
@given(data=st.data())
def test_can_generate_from_all_registered_types(data, typ):
    value = data.draw(st.from_type(typ), label="value")
    assert isinstance(value, typ)


T = TypeVar("T")


class MyGeneric(Generic[T]):
    def __init__(self, arg: T) -> None:
        self.arg = arg


class Lines(Sequence[str]):
    """Represent a sequence of text lines.

    It turns out that resolving a class which inherits from a parametrised generic
    type is... tricky.  See https://github.com/HypothesisWorks/hypothesis/issues/2951
    """


class SpecificDict(dict[int, int]):
    pass


def using_generic(instance: MyGeneric[T]) -> T:
    return instance.arg


def using_concrete_generic(instance: MyGeneric[int]) -> int:
    return instance.arg


def test_generic_origin_empty():
    with pytest.raises(ResolutionFailed):
        check_can_generate_examples(st.builds(using_generic))


def test_issue_2951_regression():
    lines_strat = st.builds(Lines, lines=st.lists(st.text()))
    prev_seq_int_repr = repr(st.from_type(Sequence[int]))
    with temp_registered(Lines, lines_strat):
        assert st.from_type(Lines) == lines_strat
        # Now let's test that the strategy for ``Sequence[int]`` did not
        # change just because we registered a strategy for ``Lines``:
        assert repr(st.from_type(Sequence[int])) == prev_seq_int_repr


def test_issue_2951_regression_two_params():
    map_strat = st.builds(SpecificDict, st.dictionaries(st.integers(), st.integers()))
    expected = repr(st.from_type(dict[int, int]))
    with temp_registered(SpecificDict, map_strat):
        assert st.from_type(SpecificDict) == map_strat
        assert expected == repr(st.from_type(dict[int, int]))


@pytest.mark.parametrize(
    "generic",
    (
        Union[str, int],
        str | int,
        Sequence[Sequence[int]],
        MyGeneric[str],
        Callable[..., str],
        Callable[[int], str],
    ),
    ids=repr,
)
@pytest.mark.parametrize("strategy", [st.none(), lambda _: st.none()])
def test_generic_origin_with_type_args(generic, strategy):
    with pytest.raises(InvalidArgument):
        st.register_type_strategy(generic, strategy)
    assert generic not in types._global_type_lookup


@pytest.mark.parametrize(
    "generic",
    (
        Callable,
        list,
        Sequence,
        # you can register types with all generic parameters
        _List[T],
        getattr(typing, "Sequence", None)[T],  # pyupgrade workaround
        list[T],
        Sequence[T],
        # User-defined generics should also work
        MyGeneric,
        MyGeneric[T],
    ),
)
def test_generic_origin_without_type_args(generic):
    with temp_registered(generic, st.just("example")):
        pass


@pytest.mark.parametrize(
    "strat, type_",
    [
        (st.from_type, MyGeneric[T]),
        (st.from_type, MyGeneric[int]),
        (st.from_type, MyGeneric),
        (st.builds, using_generic),
        (st.builds, using_concrete_generic),
    ],
    ids=get_pretty_function_description,
)
def test_generic_origin_from_type(strat, type_):
    with temp_registered(MyGeneric, st.builds(MyGeneric)):
        check_can_generate_examples(strat(type_))


def test_generic_origin_concrete_builds():
    with temp_registered(MyGeneric, st.builds(MyGeneric, st.integers())):
        assert_all_examples(
            st.builds(using_generic), lambda example: isinstance(example, int)
        )


class AbstractFoo(abc.ABC):
    def __init__(self, x):  # noqa: B027
        pass

    @abc.abstractmethod
    def qux(self):
        pass


class ConcreteFoo1(AbstractFoo):
    # Can't resolve this one due to unannotated `x` param
    def qux(self):
        pass


class ConcreteFoo2(AbstractFoo):
    def __init__(self, x: int):
        pass

    def qux(self):
        pass


@given(st.from_type(AbstractFoo))
def test_gen_abstract(foo):
    # This requires that we correctly checked which of the subclasses
    # could be resolved, rather than unconditionally using all of them.
    assert isinstance(foo, ConcreteFoo2)


class AbstractBar(abc.ABC):
    def __init__(self, x):  # noqa: B027
        pass

    @abc.abstractmethod
    def qux(self):
        pass


class ConcreteBar(AbstractBar):
    def qux(self):
        pass


def test_abstract_resolver_fallback():
    # We create our distinct strategies for abstract and concrete types
    gen_abstractbar = st.from_type(AbstractBar)
    gen_concretebar = st.builds(ConcreteBar, x=st.none())
    assert gen_abstractbar != gen_concretebar

    # And trying to generate an instance of the abstract type fails,
    # UNLESS the concrete type is currently resolvable
    with pytest.raises(ResolutionFailed):
        check_can_generate_examples(gen_abstractbar)
    with temp_registered(ConcreteBar, gen_concretebar):
        # which in turn means we resolve to the concrete subtype.
        assert_simple_property(
            gen_abstractbar, lambda gen: isinstance(gen, ConcreteBar)
        )
    with pytest.raises(ResolutionFailed):
        check_can_generate_examples(gen_abstractbar)


def _one_arg(x: int):
    assert isinstance(x, int)


def _multi_arg(x: int, y: str):
    assert isinstance(x, int)
    assert isinstance(y, str)


def _kwd_only(*, y: str):
    assert isinstance(y, str)


def _pos_and_kwd_only(x: int, *, y: str):
    assert isinstance(x, int)
    assert isinstance(y, str)


@pytest.mark.parametrize("func", [_one_arg, _multi_arg, _kwd_only, _pos_and_kwd_only])
def test_infer_all(func):
    # tests @given(...) against various signatures
    settings(max_examples=1)(given(...))(func)()


def test_does_not_add_param_empty_to_type_hints():
    def f(x):
        pass

    f.__signature__ = Signature([P("y", P.KEYWORD_ONLY)], return_annotation=None)
    assert get_type_hints(f) == {}



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from hypothesis import given, strategies as st


# Our tested class
class Product:
    def __init__(self, price: float) -> None:
        self.price: float = price

    def get_discount_price(self, discount_percentage: float):
        return self.price * (discount_percentage / 100)


# The @given decorator generates examples for us!
@given(
    price=st.floats(min_value=0, allow_nan=False, allow_infinity=False),
    discount_percentage=st.floats(
        min_value=0, max_value=100, allow_nan=False, allow_infinity=False
    ),
)
def test_a_discounted_price_is_not_higher_than_the_original_price(
    price, discount_percentage
):
    product = Product(price)
    assert product.get_discount_price(discount_percentage) <= product.price



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis.extra.array_api import (
    COMPLEX_NAMES,
    DTYPE_NAMES,
    FLOAT_NAMES,
    INT_NAMES,
    NUMERIC_NAMES,
    REAL_NAMES,
    UINT_NAMES,
)

from tests.array_api.common import MIN_VER_FOR_COMPLEX
from tests.common.debug import assert_all_examples, find_any, minimal


@pytest.mark.parametrize(
    ("strat_name", "dtype_names"),
    [
        ("integer_dtypes", INT_NAMES),
        ("unsigned_integer_dtypes", UINT_NAMES),
        ("floating_dtypes", FLOAT_NAMES),
        ("real_dtypes", REAL_NAMES),
        pytest.param(
            "complex_dtypes",
            COMPLEX_NAMES,
            marks=pytest.mark.xp_min_version(MIN_VER_FOR_COMPLEX),
        ),
    ],
)
def test_all_generated_dtypes_are_of_group(xp, xps, strat_name, dtype_names):
    """Strategy only generates expected dtypes."""
    strat_func = getattr(xps, strat_name)
    dtypes = [getattr(xp, n) for n in dtype_names]
    assert_all_examples(strat_func(), lambda dtype: dtype in dtypes)


def test_all_generated_scalar_dtypes_are_scalar(xp, xps):
    """Strategy only generates scalar dtypes."""
    if xps.api_version > "2021.12":
        dtypes = [getattr(xp, n) for n in DTYPE_NAMES]
    else:
        dtypes = [getattr(xp, n) for n in ("bool", *REAL_NAMES)]
    assert_all_examples(xps.scalar_dtypes(), lambda dtype: dtype in dtypes)


def test_all_generated_numeric_dtypes_are_numeric(xp, xps):
    """Strategy only generates numeric dtypes."""
    if xps.api_version > "2021.12":
        dtypes = [getattr(xp, n) for n in NUMERIC_NAMES]
    else:
        dtypes = [getattr(xp, n) for n in REAL_NAMES]
    assert_all_examples(xps.numeric_dtypes(), lambda dtype: dtype in dtypes)


def skipif_unsupported_complex(strat_name, dtype_name):
    if not dtype_name.startswith("complex"):
        return strat_name, dtype_name
    mark = pytest.mark.xp_min_version(MIN_VER_FOR_COMPLEX)
    return pytest.param(strat_name, dtype_name, marks=mark)


@pytest.mark.parametrize(
    ("strat_name", "dtype_name"),
    [
        *[skipif_unsupported_complex("scalar_dtypes", n) for n in DTYPE_NAMES],
        *[skipif_unsupported_complex("numeric_dtypes", n) for n in NUMERIC_NAMES],
        *[("integer_dtypes", n) for n in INT_NAMES],
        *[("unsigned_integer_dtypes", n) for n in UINT_NAMES],
        *[("floating_dtypes", n) for n in FLOAT_NAMES],
        *[("real_dtypes", n) for n in REAL_NAMES],
        *[skipif_unsupported_complex("complex_dtypes", n) for n in COMPLEX_NAMES],
    ],
)
def test_strategy_can_generate_every_dtype(xp, xps, strat_name, dtype_name):
    """Strategy generates every expected dtype."""
    strat_func = getattr(xps, strat_name)
    dtype = getattr(xp, dtype_name)
    find_any(strat_func(), lambda d: d == dtype)


def test_minimise_scalar_dtypes(xp, xps):
    """Strategy minimizes to bool dtype."""
    assert minimal(xps.scalar_dtypes()) == xp.bool


@pytest.mark.parametrize(
    "strat_name, sizes",
    [
        ("integer_dtypes", 8),
        ("unsigned_integer_dtypes", 8),
        ("floating_dtypes", 32),
        pytest.param(
            "complex_dtypes", 64, marks=pytest.mark.xp_min_version(MIN_VER_FOR_COMPLEX)
        ),
    ],
)
def test_can_specify_sizes_as_an_int(xp, xps, strat_name, sizes):
    """Strategy treats ints as a single size."""
    strat_func = getattr(xps, strat_name)
    strat = strat_func(sizes=sizes)
    find_any(strat)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import numpy as np

from hypothesis import given
from hypothesis.strategies import none


def test_numpy_prng_is_seeded():
    first = []
    prng_state = np.random.get_state()

    @given(none())
    def inner(_):
        val = np.random.bytes(10)
        if not first:
            first.append(val)
        assert val == first[0], "Numpy random module should be reproducible"

    inner()

    np.testing.assert_array_equal(
        np.random.get_state()[1], prng_state[1], "State was not restored."
    )



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import contextlib
import itertools
from random import Random

import pytest

from hypothesis import HealthCheck, Phase, settings, strategies as st
from hypothesis.database import (
    InMemoryExampleDatabase,
    choices_from_bytes,
    choices_to_bytes,
)
from hypothesis.errors import StopTest
from hypothesis.internal.conjecture.data import ConjectureData, Status
from hypothesis.internal.conjecture.engine import ConjectureRunner, RunIsComplete
from hypothesis.internal.conjecture.pareto import ParetoFront

from tests.conjecture.common import interesting_origin


def test_pareto_front_contains_different_interesting_reasons():
    def test(data):
        data.target_observations[""] = 1
        n = data.draw_integer(0, 2**4 - 1)
        data.mark_interesting(interesting_origin(n))

    runner = ConjectureRunner(
        test,
        settings=settings(
            max_examples=5000,
            database=InMemoryExampleDatabase(),
            suppress_health_check=list(HealthCheck),
        ),
        database_key=b"stuff",
        random=Random(0),
    )

    runner.run()
    assert len(runner.pareto_front) == 2**4


def test_pareto_front_omits_invalid_examples():
    def test(data):
        x = data.draw_integer(0, 2**4 - 1)
        if x % 2:
            data.target_observations[""] = 1
            data.mark_invalid()

    runner = ConjectureRunner(
        test,
        settings=settings(
            max_examples=5000,
            database=InMemoryExampleDatabase(),
            suppress_health_check=list(HealthCheck),
        ),
        database_key=b"stuff",
        random=Random(0),
    )

    runner.run()
    assert len(runner.pareto_front) == 0


def test_database_contains_only_pareto_front():
    def test(data):
        data.target_observations["1"] = data.draw(st.integers(0, 2**4))
        data.draw(st.integers(0, 2**64))
        data.target_observations["2"] = data.draw(st.integers(0, 2**8))

        assert len(set(db.fetch(b"stuff.pareto"))) == len(runner.pareto_front)

    db = InMemoryExampleDatabase()
    runner = ConjectureRunner(
        test,
        settings=settings(
            max_examples=500, database=db, suppress_health_check=list(HealthCheck)
        ),
        database_key=b"stuff",
        random=Random(0),
    )
    runner.run()

    assert len(runner.pareto_front) <= 500
    for v in runner.pareto_front:
        assert v.status >= Status.VALID

    values = set(db.fetch(b"stuff.pareto"))
    assert len(values) == len(runner.pareto_front), {
        choices_to_bytes(data.choices) for data in runner.pareto_front
    }.symmetric_difference(values)

    for data in runner.pareto_front:
        assert choices_to_bytes(data.choices) in values
        assert data in runner.pareto_front

    for b in values:
        choices = choices_from_bytes(b)
        assert runner.cached_test_function(choices) in runner.pareto_front


def test_clears_defunct_pareto_front():
    def test(data):
        data.target_observations[""] = 1
        data.draw_integer(0, 2**8 - 1)
        data.draw_integer(0, 2**8 - 1)

    db = InMemoryExampleDatabase()

    runner = ConjectureRunner(
        test,
        settings=settings(
            max_examples=10000,
            database=db,
            suppress_health_check=list(HealthCheck),
            phases=[Phase.reuse],
        ),
        database_key=b"stuff",
        random=Random(0),
    )

    for i in range(256):
        db.save(runner.pareto_key, choices_to_bytes((i, 0)))

    runner.run()
    assert len(list(db.fetch(runner.pareto_key))) == 1


def test_down_samples_the_pareto_front():
    def test(data):
        data.draw_integer(0, 2**8 - 1)
        data.draw_integer(0, 2**8 - 1)

    db = InMemoryExampleDatabase()

    runner = ConjectureRunner(
        test,
        settings=settings(
            max_examples=1000,
            database=db,
            suppress_health_check=list(HealthCheck),
            phases=[Phase.reuse],
        ),
        database_key=b"stuff",
        random=Random(0),
    )

    for n1, n2 in itertools.product(range(256), range(256)):
        db.save(runner.pareto_key, choices_to_bytes((n1, n2)))

    with pytest.raises(RunIsComplete):
        runner.reuse_existing_examples()

    assert runner.valid_examples == 1000


def test_stops_loading_pareto_front_if_interesting():
    def test(data):
        data.draw_integer()
        data.draw_integer()
        data.mark_interesting(interesting_origin())

    db = InMemoryExampleDatabase()

    runner = ConjectureRunner(
        test,
        settings=settings(
            max_examples=1000,
            database=db,
            suppress_health_check=list(HealthCheck),
            phases=[Phase.reuse],
        ),
        database_key=b"stuff",
        random=Random(0),
    )

    for n1, n2 in itertools.product(range(256), range(256)):
        db.save(runner.pareto_key, choices_to_bytes((n1, n2)))

    runner.reuse_existing_examples()
    assert runner.call_count == 1


def test_uses_tags_in_calculating_pareto_front():
    def test(data):
        data.target_observations[""] = 1
        if data.draw_boolean():
            data.start_span(11)
            data.draw_integer(0, 2**8 - 1)
            data.stop_span()

    runner = ConjectureRunner(
        test,
        settings=settings(max_examples=10, database=InMemoryExampleDatabase()),
        database_key=b"stuff",
        random=Random(0),
    )

    runner.run()
    assert len(runner.pareto_front) == 2


def test_optimises_the_pareto_front():
    def test(data):
        count = 0
        while data.draw_integer(0, 2**8 - 1):
            count += 1

        data.target_observations[""] = min(count, 5)

    runner = ConjectureRunner(
        test,
        settings=settings(max_examples=10000, database=InMemoryExampleDatabase()),
        database_key=b"stuff",
    )
    runner.cached_test_function([255] * 20 + [0])
    runner.pareto_optimise()

    assert len(runner.pareto_front) == 6
    for i, data in enumerate(runner.pareto_front):
        assert data.choices == (1,) * i + (0,)


def test_does_not_optimise_the_pareto_front_if_interesting():
    def test(data):
        n = data.draw_integer(0, 2**8 - 1)
        data.target_observations[""] = n
        if n == 255:
            data.mark_interesting(interesting_origin())

    runner = ConjectureRunner(
        test,
        settings=settings(max_examples=10000, database=InMemoryExampleDatabase()),
        database_key=b"stuff",
    )

    runner.cached_test_function([0])
    runner.pareto_optimise = None
    runner.optimise_targets()

    assert runner.interesting_examples


def test_stops_optimising_once_interesting():
    hi = 2**16 - 1

    def test(data):
        n = data.draw_integer(0, 2**16 - 1)
        data.target_observations[""] = n
        if n < hi:
            data.mark_interesting(interesting_origin())

    runner = ConjectureRunner(
        test,
        settings=settings(max_examples=10000, database=InMemoryExampleDatabase()),
        database_key=b"stuff",
    )

    data = runner.cached_test_function([hi])
    assert data.status == Status.VALID
    runner.pareto_optimise()
    assert runner.call_count <= 20
    assert runner.interesting_examples


def test_pareto_contains():
    front = ParetoFront(random=Random())
    assert "not a data" not in front

    data = ConjectureData.for_choices([])
    with contextlib.suppress(StopTest):
        data.mark_overrun()
    # check a data which turns into an Overrun when .as_result is called
    assert data not in front



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

pytest_plugins = "pytester"


TESTSUITE = """
from hypothesis import given
from hypothesis.strategies import integers

@given(integers())
def test_foo(x):
    pass

def test_bar():
    pass
"""


def test_can_select_mark(testdir):
    script = testdir.makepyfile(TESTSUITE)
    result = testdir.runpytest(
        script, "--verbose", "--strict-markers", "-m", "hypothesis"
    )
    out = "\n".join(result.stdout.lines)
    assert "1 passed, 1 deselected" in out


UNITTEST_TESTSUITE = """
from hypothesis import given
from hypothesis.strategies import integers
from unittest import TestCase

class TestStuff(TestCase):
    @given(integers())
    def test_foo(self, x):
        pass

    def test_bar(self):
        pass
"""


def test_can_select_mark_on_unittest(testdir):
    script = testdir.makepyfile(UNITTEST_TESTSUITE)
    result = testdir.runpytest(
        script, "--verbose", "--strict-markers", "-m", "hypothesis"
    )
    out = "\n".join(result.stdout.lines)
    assert "1 passed, 1 deselected" in out



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from hypothesis import given, strategies as st
from hypothesis.strategies._internal.featureflags import FeatureFlags, FeatureStrategy

from tests.common.debug import find_any, minimal

STRAT = FeatureStrategy()


def test_can_all_be_enabled():
    find_any(STRAT, lambda x: all(x.is_enabled(i) for i in range(100)))


def test_minimizes_open():
    features = range(10)

    flags = minimal(STRAT, lambda x: [x.is_enabled(i) for i in features])

    assert all(flags.is_enabled(i) for i in features)


def test_minimizes_individual_features_to_open():
    features = list(range(10))

    flags = minimal(
        STRAT, lambda x: sum(x.is_enabled(i) for i in features) < len(features)
    )

    assert all(flags.is_enabled(i) for i in features[:-1])
    assert not flags.is_enabled(features[-1])


def test_marks_unknown_features_as_enabled():
    x = find_any(STRAT, lambda v: True)

    assert x.is_enabled("fish")


def test_by_default_all_enabled():
    f = FeatureFlags()

    assert f.is_enabled("foo")


def test_eval_featureflags_repr():
    flags = FeatureFlags(enabled=["on"], disabled=["off"])
    assert flags.is_enabled("on")
    assert not flags.is_enabled("off")
    flags2 = eval(repr(flags))
    assert flags2.is_enabled("on")
    assert not flags2.is_enabled("off")


@given(st.data())
def test_repr_can_be_evalled(data):
    flags = data.draw(STRAT)

    features = data.draw(st.lists(st.text(), unique=True))

    for f in features:
        flags.is_enabled(f)

    flags2 = eval(repr(flags))

    for f in features:
        assert flags2.is_enabled(f) == flags.is_enabled(f)

    more_features = data.draw(st.lists(st.text().filter(lambda s: s not in features)))

    for f in more_features:
        assert flags2.is_enabled(f)


@given(FeatureStrategy(at_least_one_of={"a", "b", "c"}))
def test_can_avoid_disabling_every_flag(flags):
    assert any(flags.is_enabled(k) for k in {"a", "b", "c"})



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis._settings import _CI_VARS
from hypothesis.internal.compat import WINDOWS, escape_unicode_characters

pytest_plugins = "pytester"

TESTSUITE = """
from hypothesis import given, settings, Verbosity
from hypothesis.strategies import integers

@settings(verbosity=Verbosity.verbose)
@given(integers())
def test_should_be_verbose(x):
    pass

"""


@pytest.mark.parametrize("capture,expected", [("no", True), ("fd", False)])
def test_output_without_capture(testdir, capture, expected):
    script = testdir.makepyfile(TESTSUITE)
    result = testdir.runpytest(script, "--verbose", "--capture", capture)
    out = "\n".join(result.stdout.lines)
    assert "test_should_be_verbose" in out
    assert ("Trying example" in out) == expected
    assert result.ret == 0


UNICODE_EMITTING = """
import pytest
from hypothesis import given, settings, Verbosity
from hypothesis.strategies import text
import sys

def test_emits_unicode():
    @settings(verbosity=Verbosity.verbose)
    @given(text())
    def test_should_emit_unicode(t):
        assert all(ord(c) <= 1000 for c in t), ascii(t)
    with pytest.raises(AssertionError):
        test_should_emit_unicode()
"""


@pytest.mark.xfail(
    WINDOWS,
    reason="Encoding issues in running the subprocess, possibly pytest's fault",
    strict=False,  # It's possible, if rare, for this to pass on Windows too.
)
def test_output_emitting_unicode(testdir, monkeypatch):
    monkeypatch.setenv("LC_ALL", "C")
    monkeypatch.setenv("LANG", "C")
    script = testdir.makepyfile(UNICODE_EMITTING)
    result = getattr(testdir, "runpytest_subprocess", testdir.runpytest)(
        script, "--verbose", "--capture=no"
    )
    out = "\n".join(result.stdout.lines)
    assert "test_emits_unicode" in out
    assert chr(1001) in out or escape_unicode_characters(chr(1001)) in out
    assert result.ret == 0


def get_line_num(token, lines, skip_n=0):
    skipped = 0
    for i, line in enumerate(lines):
        if token in line:
            if skip_n == skipped:
                return i
            else:
                skipped += 1
    raise AssertionError(
        f"Token {token!r} not found (skipped {skipped} of planned {skip_n} skips)"
    )


TRACEBACKHIDE_HEALTHCHECK = """
from hypothesis import given, settings
from hypothesis.strategies import integers
import time
@given(integers().map(lambda x: time.sleep(0.2)))
def test_healthcheck_traceback_is_hidden(x):
    pass
"""


def test_healthcheck_traceback_is_hidden(testdir, monkeypatch):
    for key in _CI_VARS:
        monkeypatch.delenv(key, raising=False)

    script = testdir.makepyfile(TRACEBACKHIDE_HEALTHCHECK)
    lines = testdir.runpytest(script, "--verbose").stdout.lines
    def_token = "__ test_healthcheck_traceback_is_hidden __"
    timeout_token = ": FailedHealthCheck"
    def_line = get_line_num(def_token, lines)
    timeout_line = get_line_num(timeout_token, lines)
    # 16 on pytest 8.4.0 combined with py3{9, 10} or 3.13 free-threading (but
    # not with 3.13 normal??)
    assert timeout_line - def_line in {15, 16}


COMPOSITE_IS_NOT_A_TEST = """
from hypothesis.strategies import composite, none
@composite
def test_data_factory(draw):
    return draw(none())
"""


def test_deprecation_of_strategies_as_tests(testdir):
    script = testdir.makepyfile(COMPOSITE_IS_NOT_A_TEST)
    testdir.runpytest(script).assert_outcomes(failed=1)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import math
from sys import float_info

import pytest

from hypothesis import example, given, strategies as st
from hypothesis.internal.conjecture.choice import choice_equal, choice_permitted
from hypothesis.internal.conjecture.provider_conformance import float_constraints
from hypothesis.internal.floats import (
    count_between_floats,
    make_float_clamper,
    next_down,
    next_up,
    sign_aware_lte,
)

from tests.conjecture.common import float_constr


def test_can_handle_straddling_zero():
    assert count_between_floats(-0.0, 0.0) == 2


@pytest.mark.parametrize(
    "func,val",
    [
        (next_up, math.nan),
        (next_up, math.inf),
        (next_up, -0.0),
        (next_down, math.nan),
        (next_down, -math.inf),
        (next_down, 0.0),
    ],
)
def test_next_float_equal(func, val):
    if math.isnan(val):
        assert math.isnan(func(val))
    else:
        assert func(val) == val


# exponent comparisons:
@example(float_constr(1, float_info.max), 0.0)
@example(float_constr(1, float_info.max), 1.0)
@example(float_constr(1, float_info.max), 10.0)
@example(float_constr(1, float_info.max), float_info.max)
@example(float_constr(1, float_info.max), math.inf)
# mantissa comparisons:
@example(float_constr(100.0001, 100.0003), 100.0001)
@example(float_constr(100.0001, 100.0003), 100.0002)
@example(float_constr(100.0001, 100.0003), 100.0003)
@example(float_constr(100.0001, 100.0003, allow_nan=False), math.nan)
@example(float_constr(0, 10, allow_nan=False), math.nan)
@example(float_constr(0, 10, allow_nan=True), math.nan)
# the branch coverage of resampling in the "out of range of smallest magnitude" case
# relies on randomness from the mantissa. try a few different values.
@example(float_constr(-4, -1, smallest_nonzero_magnitude=4), 4.0)
@example(float_constr(-4, -1, smallest_nonzero_magnitude=4), 5.0)
@example(float_constr(-4, -1, smallest_nonzero_magnitude=4), 6.0)
@example(float_constr(1, 4, smallest_nonzero_magnitude=4), -4.0)
@example(float_constr(1, 4, smallest_nonzero_magnitude=4), -5.0)
@example(float_constr(1, 4, smallest_nonzero_magnitude=4), -6.0)
@example(float_constr(-5e-324, -0.0), 3.0)
@example(float_constr(0.0, 0.0), -0.0)
@example(float_constr(-0.0, -0.0), 0.0)
@given(float_constraints(), st.floats())
def test_float_clamper(constraints, input_value):
    min_value = constraints["min_value"]
    max_value = constraints["max_value"]
    allow_nan = constraints["allow_nan"]
    smallest_nonzero_magnitude = constraints["smallest_nonzero_magnitude"]
    clamper = make_float_clamper(
        min_value,
        max_value,
        smallest_nonzero_magnitude=smallest_nonzero_magnitude,
        allow_nan=allow_nan,
    )
    clamped = clamper(input_value)
    if math.isnan(clamped):
        # we should only clamp to nan if nans are allowed.
        assert allow_nan
    else:
        # otherwise, we should have clamped to something in the permitted range.
        assert sign_aware_lte(min_value, clamped)
        assert sign_aware_lte(clamped, max_value)

    # if input_value was permitted in the first place, then the clamped value should
    # be the same as the input value.
    if choice_permitted(
        input_value,
        {
            "min_value": min_value,
            "max_value": max_value,
            "allow_nan": allow_nan,
            "smallest_nonzero_magnitude": smallest_nonzero_magnitude,
        },
    ):
        assert choice_equal(input_value, clamped)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import numpy as np
import pandas as pd
import pytest

from hypothesis import HealthCheck, given, reject, settings, strategies as st
from hypothesis.errors import InvalidArgument
from hypothesis.extra import numpy as npst, pandas as pdst
from hypothesis.extra.pandas.impl import IntegerDtype

from tests.common.debug import (
    assert_all_examples,
    check_can_generate_examples,
    find_any,
)
from tests.pandas.helpers import supported_by_pandas


@given(pdst.data_frames([pdst.column("a", dtype=int), pdst.column("b", dtype=float)]))
def test_can_have_columns_of_distinct_types(df):
    assert df["a"].dtype == np.dtype(int)
    assert df["b"].dtype == np.dtype(float)


@given(
    pdst.data_frames(
        [pdst.column(dtype=int)], index=pdst.range_indexes(min_size=1, max_size=5)
    )
)
def test_respects_size_bounds(df):
    assert 1 <= len(df) <= 5


@given(pdst.data_frames(pdst.columns(["A", "B"], dtype=float)))
def test_can_specify_just_column_names(df):
    df["A"]
    df["B"]


@given(pdst.data_frames(pdst.columns(2, dtype=float)))
def test_can_specify_just_column_count(df):
    df[0]
    df[1]


@given(
    pdst.data_frames(
        rows=st.fixed_dictionaries({"A": st.integers(1, 10), "B": st.floats()})
    )
)
def test_gets_the_correct_data_shape_for_just_rows(table):
    assert table["A"].dtype == np.dtype("int64")
    assert table["B"].dtype == np.dtype(float)


@given(
    pdst.data_frames(
        columns=pdst.columns(["A", "B"], dtype=int),
        rows=st.lists(st.integers(0, 1000), min_size=2, max_size=2).map(sorted),
    )
)
def test_can_specify_both_rows_and_columns_list(d):
    assert d["A"].dtype == np.dtype(int)
    assert d["B"].dtype == np.dtype(int)
    for _, r in d.iterrows():
        assert r["A"] <= r["B"]


@given(
    pdst.data_frames(
        columns=pdst.columns(["A", "B"], dtype=int),
        rows=st.lists(st.integers(0, 1000), min_size=2, max_size=2)
        .map(sorted)
        .map(tuple),
    )
)
def test_can_specify_both_rows_and_columns_tuple(d):
    assert d["A"].dtype == np.dtype(int)
    assert d["B"].dtype == np.dtype(int)
    for _, r in d.iterrows():
        assert r["A"] <= r["B"]


@given(
    pdst.data_frames(
        columns=pdst.columns(["A", "B"], dtype=int),
        rows=st.lists(st.integers(0, 1000), min_size=2, max_size=2).map(
            lambda x: {"A": min(x), "B": max(x)}
        ),
    )
)
def test_can_specify_both_rows_and_columns_dict(d):
    assert d["A"].dtype == np.dtype(int)
    assert d["B"].dtype == np.dtype(int)
    for _, r in d.iterrows():
        assert r["A"] <= r["B"]


@given(
    pdst.data_frames(
        [
            pdst.column(
                "A",
                fill=st.just(np.nan),
                dtype=float,
                elements=st.floats(allow_nan=False),
            )
        ],
        rows=st.builds(dict),
    )
)
def test_can_fill_in_missing_elements_from_dict(df):
    assert np.isnan(df["A"]).all()


@st.composite
def column_strategy(draw):
    name = draw(st.none() | st.text())
    dtype = draw(npst.scalar_dtypes().filter(supported_by_pandas))
    pass_dtype = not draw(st.booleans())
    if pass_dtype:
        pass_elements = not draw(st.booleans())
    else:
        pass_elements = True
    if pass_elements:
        elements = npst.from_dtype(dtype)
    else:
        elements = None

    unique = draw(st.booleans())
    fill = st.nothing() if draw(st.booleans()) else None

    return pdst.column(
        name=name, dtype=dtype, unique=unique, fill=fill, elements=elements
    )


@given(pdst.data_frames(pdst.columns(1, dtype=np.dtype("M8[ns]"))))
def test_data_frames_with_timestamp_columns(df):
    pass


@given(
    pdst.data_frames(
        pdst.columns(["A"], dtype=float, fill=st.just(np.nan), unique=True)
    )
)
def test_unique_column_with_fill(df):
    assert len(set(df["A"])) == len(df["A"])


@settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
@given(st.data())
def test_arbitrary_data_frames(data):
    columns = data.draw(
        st.lists(
            column_strategy(),
            unique_by=lambda c: c.name if c.name is not None else np.nan,
        )
    )

    try:
        # Use raw data to work around pandas bug in repr. See
        # https://github.com/pandas-dev/pandas/issues/27484
        df = data.conjecture_data.draw(pdst.data_frames(columns))
    except Exception as e:
        if type(e).__name__ == "OutOfBoundsDatetime":
            # See https://github.com/HypothesisWorks/hypothesis-python/pull/826
            reject()
        else:
            raise
    data_frame_columns = list(df)

    assert len(data_frame_columns) == len(columns)

    for i, (c, n) in enumerate(zip(columns, df, strict=True)):
        if c.name is None:
            assert n == i
        else:
            assert c.name == n

    for i, c in enumerate(columns):
        column_name = data_frame_columns[i]
        values = df[column_name]
        if c.unique:
            # NA values should always be treated as unique to each other, so we
            # just ignore them here. Note NA values in the ecosystem can have
            # different identity behaviours, e.g.
            #
            #     >>> set([float("nan"), float("nan")])
            #     {nan, nan}
            #     >>> set([pd.NaT, pd.NaT])
            #     {NaT}
            #
            non_na_values = values.dropna()
            assert len(set(non_na_values)) == len(non_na_values)


@given(
    pdst.data_frames(
        pdst.columns(["A"], unique=True, dtype=int), rows=st.tuples(st.integers(0, 10))
    )
)
def test_can_specify_unique_with_rows(df):
    column = df["A"]
    assert len(set(column)) == len(column)


def test_uniqueness_does_not_affect_other_rows_1():
    data_frames = pdst.data_frames(
        [
            pdst.column("A", dtype=int, unique=True),
            pdst.column("B", dtype=int, unique=False),
        ],
        rows=st.tuples(st.integers(0, 10), st.integers(0, 10)),
        index=pdst.range_indexes(2, 2),
    )
    find_any(data_frames, lambda x: x["B"][0] == x["B"][1])


def test_uniqueness_does_not_affect_other_rows_2():
    data_frames = pdst.data_frames(
        [
            pdst.column("A", dtype=bool, unique=False),
            pdst.column("B", dtype=int, unique=True),
        ],
        rows=st.tuples(st.booleans(), st.integers(0, 10)),
        index=pdst.range_indexes(2, 2),
    )
    find_any(data_frames, lambda x: x["A"][0] == x["A"][1])


@given(
    pdst.data_frames(pdst.columns(["A"], dtype=int, fill=st.just(7)), rows=st.tuples())
)
def test_will_fill_missing_columns_in_tuple_row(df):
    for d in df["A"]:
        assert d == 7


@settings(suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow])
@given(
    pdst.data_frames(
        index=pdst.range_indexes(10, 10),
        columns=[pdst.column(elements=st.integers(0, 9), fill=None, unique=True)],
    )
)
def test_can_generate_unique_columns(df):
    assert set(df[0]) == set(range(10))


@pytest.mark.skip(reason="Just works on Pandas 1.4, though the changelog is silent")
@pytest.mark.parametrize("dtype", [None, object])
def test_expected_failure_from_omitted_object_dtype(dtype):
    # See https://github.com/HypothesisWorks/hypothesis/issues/3133
    col = pdst.column(elements=st.sets(st.text(), min_size=1), dtype=dtype)

    @given(pdst.data_frames(columns=[col]))
    def works_with_object_dtype(df):
        pass

    if dtype is object:
        works_with_object_dtype()
    else:
        assert dtype is None
        with pytest.raises(ValueError, match="Maybe passing dtype=object would help"):
            works_with_object_dtype()


@pytest.mark.skipif(
    not IntegerDtype, reason="Nullable types not available in this version of Pandas"
)
def test_pandas_nullable_types():
    st = pdst.data_frames(pdst.columns(2, dtype=pd.core.arrays.integer.Int8Dtype()))
    df = find_any(st, lambda s: s.isna().any().any())
    for s in df.columns:
        assert type(df[s].dtype) == pd.core.arrays.integer.Int8Dtype


@given(pdst.data_frames(columns=[pdst.column("col", dtype=object)]))
def test_object_columns_are_of_type_object(df):
    assert df["col"].dtype == np.dtype("O")


def test_class_instances_not_allowed_in_scalar_columns():
    class A:
        pass

    s = pdst.data_frames(
        columns=[
            pdst.column(
                "col",
                elements=st.just(A()),
                dtype=npst.scalar_dtypes(),
            )
        ]
    )

    with pytest.raises(InvalidArgument):
        check_can_generate_examples(s)


def test_can_generate_object_arrays_with_mixed_dtype_elements():
    class A:
        pass

    s = pdst.data_frames(
        columns=[pdst.column("col", st.just(A()) | st.integers(), dtype=object)],
        index=pdst.range_indexes(1),
    )
    assert_all_examples(s, lambda df: df["col"].dtype == np.dtype("O"))
    find_any(s, lambda df: len({type(x) for x in df["col"].values}) > 1)


@given(st.data())
@pytest.mark.xfail(
    strict=False,
    reason="not actually true due to pandas conversion. see "
    "https://github.com/HypothesisWorks/hypothesis/pull/4444#issuecomment-3413951478",
)
def test_object_dataframe_can_hold_arbitrary_class_instances(data):
    instance = data.draw(st.from_type(type).flatmap(st.from_type))
    s = pdst.data_frames(
        columns=[pdst.column("col", elements=st.just(instance), dtype=object)]
    )
    df = data.draw(s)

    assert all(v is instance for v in df["col"].values)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import math

import pytest

from hypothesis import given, settings, strategies as st
from hypothesis.errors import InvalidArgument
from hypothesis.extra.array_api import COMPLEX_NAMES, REAL_NAMES
from hypothesis.internal.floats import width_smallest_normals

from tests.array_api.common import (
    MIN_VER_FOR_COMPLEX,
    dtype_name_params,
    flushes_to_zero,
)
from tests.common.debug import (
    assert_all_examples,
    check_can_generate_examples,
    find_any,
    minimal,
)
from tests.common.utils import flaky


def skip_on_missing_unique_values(xp):
    if not hasattr(xp, "unique_values"):
        pytest.mark.skip("xp.unique_values() is not required to exist")


def xfail_on_indistinct_nans(xp):
    """
    xp.unique_value() should return distinct NaNs - if not, tests that (rightly)
    assume such behaviour will likely fail. For example, NumPy 1.22 treats NaNs
    as indistinct, so tests that use this function will be marked as xfail.
    See https://mail.python.org/pipermail/numpy-discussion/2021-August/081995.html
    """
    skip_on_missing_unique_values(xp)
    two_nans = xp.asarray([float("nan"), float("nan")])
    if xp.unique_values(two_nans).size != 2:
        pytest.xfail("NaNs not distinct")


@pytest.mark.parametrize("dtype_name", dtype_name_params)
def test_draw_arrays_from_dtype(xp, xps, dtype_name):
    """Draw arrays from dtypes."""
    dtype = getattr(xp, dtype_name)
    assert_all_examples(xps.arrays(dtype, ()), lambda x: x.dtype == dtype)


@pytest.mark.parametrize("dtype_name", dtype_name_params)
def test_draw_arrays_from_scalar_names(xp, xps, dtype_name):
    """Draw arrays from dtype names."""
    dtype = getattr(xp, dtype_name)
    assert_all_examples(xps.arrays(dtype_name, ()), lambda x: x.dtype == dtype)


@given(data=st.data())
def test_draw_arrays_from_shapes(xp, xps, data):
    """Draw arrays from shapes."""
    shape = data.draw(xps.array_shapes())
    x = data.draw(xps.arrays(xp.int8, shape))
    assert x.ndim == len(shape)
    assert x.shape == shape


@given(data=st.data())
def test_draw_arrays_from_int_shapes(xp, xps, data):
    """Draw arrays from integers as shapes."""
    size = data.draw(st.integers(0, 10))
    x = data.draw(xps.arrays(xp.int8, size))
    assert x.shape == (size,)


@pytest.mark.parametrize(
    "strat_name",
    [
        "scalar_dtypes",
        "boolean_dtypes",
        "integer_dtypes",
        "unsigned_integer_dtypes",
        "floating_dtypes",
        "real_dtypes",
        pytest.param(
            "complex_dtypes", marks=pytest.mark.xp_min_version(MIN_VER_FOR_COMPLEX)
        ),
    ],
)
def test_draw_arrays_from_dtype_strategies(xp, xps, strat_name):
    """Draw arrays from dtype strategies."""
    strat_func = getattr(xps, strat_name)
    strat = strat_func()
    find_any(xps.arrays(strat, ()))


@settings(deadline=None)
@given(data=st.data())
def test_draw_arrays_from_dtype_name_strategies(xp, xps, data):
    """Draw arrays from dtype name strategies."""
    all_names = ("bool", *REAL_NAMES)
    if xps.api_version > "2021.12":
        all_names += COMPLEX_NAMES
    sample_names = data.draw(
        st.lists(st.sampled_from(all_names), min_size=1, unique=True)
    )
    find_any(xps.arrays(st.sampled_from(sample_names), ()))


def test_generate_arrays_from_shapes_strategy(xp, xps):
    """Generate arrays from shapes strategy."""
    find_any(xps.arrays(xp.int8, xps.array_shapes()))


def test_generate_arrays_from_integers_strategy_as_shape(xp, xps):
    """Generate arrays from integers strategy as shapes strategy."""
    find_any(xps.arrays(xp.int8, st.integers(0, 100)))


def test_generate_arrays_from_zero_dimensions(xp, xps):
    """Generate arrays from empty shape."""
    assert_all_examples(xps.arrays(xp.int8, ()), lambda x: x.shape == ())


@given(data=st.data())
def test_generate_arrays_from_zero_sided_shapes(xp, xps, data):
    """Generate arrays from shapes with at least one 0-sized dimension."""
    shape = data.draw(xps.array_shapes(min_side=0).filter(lambda s: 0 in s))
    arr = data.draw(xps.arrays(xp.int8, shape))
    assert arr.shape == shape


def test_generate_arrays_from_unsigned_ints(xp, xps):
    """Generate arrays from unsigned integer dtype."""
    assert_all_examples(xps.arrays(xp.uint32, (5, 5)), lambda x: xp.all(x >= 0))
    # Ensure we're not just picking non-negative signed integers
    signed_max = xp.iinfo(xp.int32).max
    find_any(xps.arrays(xp.uint32, (5, 5)), lambda x: xp.any(x > signed_max))


def test_generate_arrays_from_0d_arrays(xp, xps):
    """Generate arrays from 0d array elements."""
    assert_all_examples(
        xps.arrays(
            dtype=xp.uint8,
            shape=(5, 5),
            elements=xps.from_dtype(xp.uint8).map(
                lambda e: xp.asarray(e, dtype=xp.uint8)
            ),
        ),
        lambda x: x.shape == (5, 5),
    )


def test_minimize_arrays_with_default_dtype_shape_strategies(xp, xps):
    """Strategy with default scalar_dtypes and array_shapes strategies minimize
    to a boolean 1-dimensional array of size 1."""
    smallest = minimal(xps.arrays(xps.scalar_dtypes(), xps.array_shapes()))
    assert smallest.shape == (1,)
    assert smallest.dtype == xp.bool
    assert not xp.any(smallest)


def test_minimize_arrays_with_0d_shape_strategy(xp, xps):
    """Strategy with shape strategy that can generate empty tuples minimizes to
    0d arrays."""
    smallest = minimal(xps.arrays(xp.int8, xps.array_shapes(min_dims=0)))
    assert smallest.shape == ()


@pytest.mark.parametrize("dtype", dtype_name_params[1:])
def test_minimizes_numeric_arrays(xp, xps, dtype):
    """Strategies with numeric dtypes minimize to zero-filled arrays."""
    smallest = minimal(xps.arrays(dtype, (2, 2)))
    assert xp.all(smallest == 0)


def test_minimize_large_uint_arrays(xp, xps):
    """Strategy with uint dtype and largely sized shape minimizes to a good
    example."""
    if not hasattr(xp, "nonzero"):
        pytest.skip("optional API")
    smallest = minimal(xps.arrays(xp.uint8, 100), lambda x: xp.any(x) and not xp.all(x))
    assert xp.all(xp.logical_or(smallest == 0, smallest == 1))
    idx = xp.nonzero(smallest)[0]
    assert idx.size in (1, smallest.size - 1)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@flaky(max_runs=50, min_passes=1)
def test_minimize_float_arrays(xp, xps):
    """Strategy with float dtype minimizes to a good example.

    We filter runtime warnings and expect flaky array generation for
    specifically NumPy - this behaviour may not be required when testing
    with other array libraries.
    """
    smallest = minimal(xps.arrays(xp.float32, 50), lambda x: xp.sum(x) >= 1.0)
    # TODO_IR the shrinker gets stuck when the first failure is math.inf, because
    # downcasting inf to a float32 overflows, triggering rejection sampling which
    # is then immediately not a shrink (specifically it overruns the attempt data).
    #
    # this should be resolved by adding float widths to the ir.
    assert xp.sum(smallest) in (1, 50) or all(math.isinf(v) for v in smallest)


def test_minimizes_to_fill(xp, xps):
    """Strategy with single fill value minimizes to arrays only containing said
    fill value."""
    smallest = minimal(xps.arrays(xp.float32, 10, fill=st.just(3.0)))
    assert xp.all(smallest == 3.0)


def test_generate_unique_arrays(xp, xps):
    """Generates unique arrays."""
    skip_on_missing_unique_values(xp)
    assert_all_examples(
        xps.arrays(xp.int8, st.integers(0, 20), unique=True),
        lambda x: xp.unique_values(x).size == x.size,
    )


def test_cannot_draw_unique_arrays_with_too_small_elements(xp, xps):
    """Unique strategy with elements strategy range smaller than its size raises
    helpful error."""
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(
            xps.arrays(xp.int8, 10, elements=st.integers(0, 5), unique=True)
        )


def test_cannot_fill_arrays_with_non_castable_value(xp, xps):
    """Strategy with fill not castable to dtype raises helpful error."""
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(
            xps.arrays(xp.int8, 10, fill=st.just("not a castable value"))
        )


def test_generate_unique_arrays_with_high_collision_elements(xp, xps):
    """Generates unique arrays with just elements of 0.0 and NaN fill."""

    @given(
        xps.arrays(
            dtype=xp.float32,
            shape=st.integers(0, 20),
            elements=st.just(0.0),
            fill=st.just(xp.nan),
            unique=True,
        )
    )
    def test(x):
        zero_mask = x == 0.0
        assert xp.sum(xp.astype(zero_mask, xp.uint8)) <= 1

    test()


def test_generate_unique_arrays_using_all_elements(xp, xps):
    """Unique strategy with elements strategy range equal to its size will only
    generate arrays with one of each possible element."""
    skip_on_missing_unique_values(xp)
    assert_all_examples(
        xps.arrays(xp.int8, (4,), elements=st.integers(0, 3), unique=True),
        lambda x: xp.unique_values(x).size == x.size,
    )


def test_may_fill_unique_arrays_with_nan(xp, xps):
    """Unique strategy with NaN fill can generate arrays holding NaNs."""
    find_any(
        xps.arrays(
            dtype=xp.float32,
            shape=10,
            elements={"allow_nan": False},
            unique=True,
            fill=st.just(xp.nan),
        ),
        lambda x: xp.any(xp.isnan(x)),
    )


def test_may_not_fill_unique_array_with_non_nan(xp, xps):
    """Unique strategy with just fill elements of 0.0 raises helpful error."""
    strat = xps.arrays(
        dtype=xp.float32,
        shape=10,
        elements={"allow_nan": False},
        unique=True,
        fill=st.just(0.0),
    )
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(strat)


def test_floating_point_array():
    import warnings

    from hypothesis.extra.array_api import make_strategies_namespace

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            import numpy.array_api as nxp
    except ModuleNotFoundError:
        import numpy as nxp
    xps = make_strategies_namespace(nxp)
    dtypes = xps.floating_dtypes() | xps.complex_dtypes()

    strat = xps.arrays(dtype=dtypes, shape=10)

    check_can_generate_examples(strat)


@pytest.mark.parametrize(
    "kwargs",
    [
        {"elements": st.just(300)},
        {"elements": st.nothing(), "fill": st.just(300)},
    ],
)
def test_may_not_use_overflowing_integers(xp, xps, kwargs):
    """Strategy with elements strategy range outside the dtype's bounds raises
    helpful error."""
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(xps.arrays(dtype=xp.int8, shape=1, **kwargs))


@pytest.mark.parametrize("fill", [False, True])
@pytest.mark.parametrize(
    "dtype, strat",
    [
        ("float32", st.floats(min_value=10**40, allow_infinity=False)),
        ("float64", st.floats(min_value=10**40, allow_infinity=False)),
        pytest.param(
            "complex64",
            st.complex_numbers(min_magnitude=10**300, allow_infinity=False),
            marks=pytest.mark.xp_min_version(MIN_VER_FOR_COMPLEX),
        ),
    ],
)
def test_may_not_use_unrepresentable_elements(xp, xps, fill, dtype, strat):
    """Strategy with elements not representable by the dtype raises helpful error."""
    if fill:
        kw = {"elements": st.nothing(), "fill": strat}
    else:
        kw = {"elements": strat}
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(xps.arrays(dtype=dtype, shape=1, **kw))


def test_floats_can_be_constrained(xp, xps):
    """Strategy with float dtype and specified elements strategy range
    (inclusive) generates arrays with elements inside said range."""
    assert_all_examples(
        xps.arrays(
            dtype=xp.float32, shape=10, elements={"min_value": 0, "max_value": 1}
        ),
        lambda x: xp.all(x >= 0) and xp.all(x <= 1),
    )


def test_floats_can_be_constrained_excluding_endpoints(xp, xps):
    """Strategy with float dtype and specified elements strategy range
    (exclusive) generates arrays with elements inside said range."""
    assert_all_examples(
        xps.arrays(
            dtype=xp.float32,
            shape=10,
            elements={
                "min_value": 0,
                "max_value": 1,
                "exclude_min": True,
                "exclude_max": True,
            },
        ),
        lambda x: xp.all(x > 0) and xp.all(x < 1),
    )


def test_is_still_unique_with_nan_fill(xp, xps):
    """Unique strategy with NaN fill generates unique arrays."""
    skip_on_missing_unique_values(xp)
    xfail_on_indistinct_nans(xp)
    assert_all_examples(
        xps.arrays(
            dtype=xp.float32,
            elements={"allow_nan": False},
            shape=10,
            unique=True,
            fill=st.just(xp.nan),
        ),
        lambda x: xp.unique_values(x).size == x.size,
    )


def test_unique_array_with_fill_can_use_all_elements(xp, xps):
    """Unique strategy with elements range equivalent to its size and NaN fill
    can generate arrays with all possible values."""
    skip_on_missing_unique_values(xp)
    xfail_on_indistinct_nans(xp)
    find_any(
        xps.arrays(
            dtype=xp.float32,
            shape=10,
            unique=True,
            elements=st.integers(1, 9),
            fill=st.just(xp.nan),
        ),
        lambda x: xp.unique_values(x).size == x.size,
    )


def test_generate_unique_arrays_without_fill(xp, xps):
    """Generate arrays from unique strategy with no fill.

    Covers the collision-related branches for fully dense unique arrays.
    Choosing 25 of 256 possible values means we're almost certain to see
    collisions thanks to the birthday paradox, but finding unique values should
    still be easy.
    """
    skip_on_missing_unique_values(xp)
    assert_all_examples(
        xps.arrays(dtype=xp.uint8, shape=25, unique=True, fill=st.nothing()),
        lambda x: xp.unique_values(x).size == x.size,
    )


def test_efficiently_generate_unique_arrays_using_all_elements(xp, xps):
    """Unique strategy with elements strategy range equivalent to its size
    generates arrays with all possible values. Generation is not too slow.

    Avoids the birthday paradox with UniqueSampledListStrategy.
    """
    skip_on_missing_unique_values(xp)
    assert_all_examples(
        xps.arrays(dtype=xp.int8, shape=255, unique=True),
        lambda x: xp.unique_values(x).size == x.size,
    )


@given(st.data(), st.integers(-100, 100), st.integers(1, 100))
def test_array_element_rewriting(xp, xps, data, start, size):
    """Unique strategy generates arrays with expected elements."""
    x = data.draw(
        xps.arrays(
            dtype=xp.int64,
            shape=size,
            elements=st.integers(start, start + size - 1),
            unique=True,
        )
    )
    x_set_expect = xp.arange(start, start + size, dtype=xp.int64)
    x_set = xp.sort(xp.unique_values(x))
    assert xp.all(x_set == x_set_expect)


def test_generate_0d_arrays_with_no_fill(xp, xps):
    """Generate arrays with zero-dimensions and no fill."""
    assert_all_examples(
        xps.arrays(xp.bool, (), fill=st.nothing()),
        lambda x: x.dtype == xp.bool and x.shape == (),
    )


@pytest.mark.parametrize("dtype", ["float32", "float64"])
@pytest.mark.parametrize("low", [-2.0, -1.0, 0.0, 1.0])
@given(st.data())
def test_excluded_min_in_float_arrays(xp, xps, dtype, low, data):
    """Strategy with elements strategy excluding min does not generate arrays
    with elements less or equal to said min."""
    strat = xps.arrays(
        dtype=dtype,
        shape=(),
        elements={
            "min_value": low,
            "max_value": low + 1,
            "exclude_min": True,
        },
    )
    x = data.draw(strat, label="array")
    assert xp.all(x > low)


@st.composite
def distinct_int64_integers(draw):
    used = draw(st.shared(st.builds(set), key="distinct_int64_integers.used"))
    i = draw(st.integers(-(2**63), 2**63 - 1).filter(lambda x: x not in used))
    used.add(i)
    return i


def test_does_not_reuse_distinct_integers(xp, xps):
    """Strategy with distinct integer elements strategy generates arrays with
    distinct values."""
    skip_on_missing_unique_values(xp)
    assert_all_examples(
        xps.arrays(xp.int64, 10, elements=distinct_int64_integers()),
        lambda x: xp.unique_values(x).size == x.size,
    )


def test_may_reuse_distinct_integers_if_asked(xp, xps):
    """Strategy with shared elements and fill strategies of distinct integers
    may generate arrays with non-distinct values."""
    skip_on_missing_unique_values(xp)
    find_any(
        xps.arrays(
            xp.int64,
            10,
            elements=distinct_int64_integers(),
            fill=distinct_int64_integers(),
        ),
        lambda x: xp.unique_values(x).size < x.size,
    )


def test_subnormal_elements_validation(xp, xps):
    """Strategy with subnormal elements strategy is correctly validated.

    For FTZ builds of array modules, a helpful error should raise. Conversely,
    for builds of array modules which support subnormals, the strategy should
    generate arrays without raising.
    """
    elements = {
        "min_value": 0.0,
        "max_value": width_smallest_normals[32],
        "exclude_min": True,
        "exclude_max": True,
        "allow_subnormal": True,
    }
    strat = xps.arrays(xp.float32, 10, elements=elements)
    if flushes_to_zero(xp, width=32):
        with pytest.raises(InvalidArgument, match="Generated subnormal float"):
            check_can_generate_examples(strat)
    else:
        check_can_generate_examples(strat)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import os
import shlex
import subprocess
from pathlib import Path


def current_branch():
    return (
        subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"])
        .decode("ascii")
        .strip()
    )


def tags():
    result = [
        t.decode("ascii") for t in subprocess.check_output(["git", "tag"]).split(b"\n")
    ]
    assert len(set(result)) == len(result)
    return set(result)


ROOT = Path(
    subprocess.check_output(
        ["git", "-C", os.path.dirname(__file__), "rev-parse", "--show-toplevel"]
    )
    .decode("ascii")
    .strip()
)
REPO_TESTS = ROOT / "whole_repo_tests"


def hash_for_name(name):
    return subprocess.check_output(["git", "rev-parse", name]).decode("ascii").strip()


def is_ancestor(a, b):
    check = subprocess.call(["git", "merge-base", "--is-ancestor", a, b])
    assert 0 <= check <= 1
    return check == 0


def merge_base(a, b):
    return subprocess.check_output(["git", "merge-base", a, b]).strip()


def point_of_divergence():
    return merge_base("HEAD", "origin/master")


def has_changes(files):
    command = [
        "git",
        "diff",
        "--no-patch",
        "--exit-code",
        point_of_divergence(),
        "HEAD",
        "--",
        *files,
    ]
    return subprocess.call(command) != 0


def has_uncommitted_changes(filename):
    return subprocess.call(["git", "diff", "--exit-code", filename]) != 0


def last_committer():
    out, _ = subprocess.Popen(
        ["git", "log", "-1", "--pretty=format:%an"],
        stdout=subprocess.PIPE,
        universal_newlines=True,
    ).communicate()
    return out


def git(*args):
    subprocess.check_call(("git", *args))


TOOLING_COMMITTER_NAME = "CI on behalf of the Hypothesis team"


def configure_git():
    git("config", "user.name", TOOLING_COMMITTER_NAME)
    git("config", "user.email", "david@drmaciver.com")


def create_tag(tagname):
    assert tagname not in tags()
    git("tag", tagname)


def push_tag(tagname):
    assert_can_release()
    subprocess.check_call(["git", "push", "origin", shlex.quote(tagname)])
    subprocess.check_call(["git", "push", "origin", "HEAD:master"])


def assert_can_release():
    assert not IS_PULL_REQUEST, "Cannot release from pull requests"


def modified_files():
    files = set()
    for command in [
        [
            "git",
            "diff",
            "--name-only",
            "--diff-filter=d",
            point_of_divergence(),
            "HEAD",
        ],
        ["git", "diff", "--name-only"],
    ]:
        diff_output = subprocess.check_output(command).decode("ascii")
        for l in diff_output.split("\n"):
            filepath = l.strip()
            if filepath and os.path.exists(filepath):
                files.add(Path(filepath))
    return files


def all_files() -> list[Path]:
    return [
        Path(f)
        for f in subprocess.check_output(["git", "ls-files"])
        .decode("ascii")
        .splitlines()
        if os.path.exists(f)
    ]


def changed_files_from_master():
    """Returns a list of files which have changed between a branch and master."""
    files = set()
    command = ["git", "diff", "--name-only", "HEAD", "master"]
    diff_output = subprocess.check_output(command).decode("ascii")
    for line in diff_output.splitlines():
        filepath = line.strip()
        if filepath:
            files.add(filepath)
    return files


IS_PULL_REQUEST = os.environ.get("GITHUB_REF", "").startswith("refs/pull/")


def all_projects():
    import hypothesistooling.projects.hypothesispython as hp

    return [hp]



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import os
import subprocess
import sys

from tests.common.utils import skipif_emscripten

SHOULD_NOT_IMPORT_TEST_RUNNERS = """
import sys
import unittest
from hypothesis import given, strategies as st

class TestDoesNotImportRunners(unittest.TestCase):
    strat = st.integers() | st.floats() | st.sampled_from(["a", "b"])

    @given(strat)
    def test_does_not_import_unittest2(self, x):
        assert "unittest2" not in sys.modules

    @given(strat)
    def test_does_not_import_nose(self, x):
        assert "nose" not in sys.modules
        assert "nose2" not in sys.modules

    @given(strat)
    def test_does_not_import_pytest(self, x):
        assert "pytest" not in sys.modules

if __name__ == '__main__':
    unittest.main()
"""


@skipif_emscripten
def test_hypothesis_does_not_import_test_runners(tmp_path):
    # We obviously can't use pytest to check that pytest is not imported,
    # so for consistency we use unittest for all three non-stdlib test runners.
    # It's unclear which of our dependencies is importing unittest, but
    # since I doubt it's causing any spurious failures I don't really care.
    # See https://github.com/HypothesisWorks/hypothesis/pull/2204
    fname = tmp_path / "test.py"
    fname.write_text(SHOULD_NOT_IMPORT_TEST_RUNNERS, encoding="utf-8")
    subprocess.check_call(
        [sys.executable, str(fname)],
        env={**os.environ, "HYPOTHESIS_NO_PLUGINS": "1"},
    )



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Try running `hypothesis write ...` on all available modules.

Prints a list of module names which caused some kind of internal error.

The idea here is to check that we at least don't crash on anything that
people actually ship, or at least only on the cases we know and don't
really care about - there are a lot of strange things in a python install.
Some have import-time side effects or errors so we skip them; others
just have such weird semantics that we don't _want_ to support them.
"""

import multiprocessing
import os
import subprocess
import sysconfig

skip = (
    "idlelib curses antigravity pip prompt_toolkit IPython .popen_ django. .test. "
    "execnet.script lib2to3.pgen2.conv tests. Cython. ~ - ._ libcst.codemod. "
    "modernize flask. sphinx. pyasn1 dbm.ndbm doctest"
).split()


def getmodules():
    std_lib = sysconfig.get_python_lib(standard_lib=True)
    for top, _, files in os.walk(std_lib):
        for nm in files:
            if nm.endswith(".py") and nm not in ("__init__.py", "__main__.py"):
                modname = (
                    os.path.join(top, nm)[len(std_lib) + 1 : -3]
                    .replace(os.sep, ".")
                    .replace("site-packages.", "")
                )
                if not any(bad in modname for bad in skip):
                    yield modname


def write_for(mod):
    try:
        subprocess.run(
            ["hypothesis", "write", mod],
            check=True,
            capture_output=True,
            timeout=10,
            text=True,
            encoding="utf-8",
        )
    except subprocess.SubprocessError as e:
        # Only report the error if we could load _but not process_ the module
        if (
            "Error: Found the '" not in e.stderr
            and "Error: Failed to import" not in e.stderr
        ):
            return mod


if __name__ == "__main__":
    print("# prints the names of modules for which `hypothesis write` errors out")
    with multiprocessing.Pool() as pool:
        for name in pool.imap(write_for, getmodules()):
            if name is not None:
                print(name, flush=True)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from unittest.mock import Mock, create_autospec

import pytest

from hypothesis import example, given
from hypothesis.strategies import integers

from tests.common.utils import fails, skipif_threading

pytest_plugins = "pytester"


@pytest.fixture(scope="session")
def infinity():
    return float("inf")


@pytest.fixture(scope="module")
def mock_fixture():
    return Mock()


@pytest.fixture(scope="module")
def spec_fixture():
    class Foo:
        def __init__(self):
            pass

        def bar(self):
            return "baz"

    return create_autospec(Foo)


@given(integers())
def test_can_mix_fixture_and_positional_strategy(infinity, xs):
    # Hypothesis fills arguments from the right, so if @given() uses
    # positional arguments then any strategies need to be on the right.
    assert xs <= infinity


@given(xs=integers())
def test_can_mix_fixture_and_keyword_strategy(xs, infinity):
    assert xs <= infinity


@example(xs=0)
@given(xs=integers())
def test_can_mix_fixture_example_and_keyword_strategy(xs, infinity):
    assert xs <= infinity


@fails
@given(integers())
def test_can_inject_mock_via_fixture(mock_fixture, xs):
    """A negative test is better for this one - this condition uncovers a bug
    whereby the mock fixture is executed instead of the test body and always
    succeeds. If this test fails, then we know we've run the test body instead
    of the mock.
    """
    raise AssertionError


@given(integers())
def test_can_inject_autospecced_mock_via_fixture(spec_fixture, xs):
    spec_fixture.bar.return_value = float("inf")
    assert xs <= spec_fixture.bar()


TESTSUITE = """
import pytest
from hypothesis import given, strategies as st

@pytest.fixture(scope="function", autouse=True)
def autofix(request):
    pass

@given(x=st.integers())
def test_requests_function_scoped_fixture(capsys, x):
    pass

@pytest.mark.parametrize("percent", ["%", "%s"])
@given(x=st.integers())
def test_requests_function_scoped_fixture_percent_parametrized(capsys, x, percent):
    # See https://github.com/HypothesisWorks/hypothesis/issues/2469
    pass

class TestClass:
    @given(x=st.integers())
    def test_requests_method_scoped_fixture(capsys, x):
        pass

@given(x=st.integers())
def test_autouse_function_scoped_fixture(x):
    pass
"""


def test_given_plus_function_scoped_non_autouse_fixtures_are_deprecated(testdir):
    script = testdir.makepyfile(TESTSUITE)
    testdir.runpytest(script).assert_outcomes(passed=1, failed=4)


CONFTEST_SUPPRESS = """
from hypothesis import HealthCheck, settings

settings.register_profile(
    "suppress",
    suppress_health_check=[HealthCheck.function_scoped_fixture],
)
"""


@skipif_threading
def test_suppress_fixture_health_check_via_profile(testdir):
    script = testdir.makepyfile(TESTSUITE)
    testdir.makeconftest(CONFTEST_SUPPRESS)

    testdir.runpytest(script).assert_outcomes(passed=1, failed=4)
    testdir.runpytest(script, "--hypothesis-profile=suppress").assert_outcomes(passed=5)


TESTSCRIPT_SUPPRESS_FIXTURE = """
import pytest
from hypothesis import HealthCheck, given, settings, strategies as st

@given(x=st.integers())
def test_fails_health_check(capsys, x):
    pass

@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(x=st.integers())
def test_suppresses_health_check(capsys, x):
    pass

@given(x=st.integers())
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_suppresses_health_check_2(capsys, x):
    pass
"""


def test_suppress_health_check_function_scoped_fixture(testdir):
    script = testdir.makepyfile(TESTSCRIPT_SUPPRESS_FIXTURE)
    testdir.runpytest(script).assert_outcomes(passed=2, failed=1)


TESTSCRIPT_OVERRIDE_FIXTURE = """
import pytest
from hypothesis import given, strategies as st

@pytest.fixture(scope="function", name="event_loop")
def event_loop_1():
    return

@pytest.fixture(scope="module", name="event_loop")
def event_loop_2():
    return

@given(x=st.integers())
def test_override_fixture(event_loop, x):
    pass
"""


def test_given_plus_overridden_fixture(testdir):
    script = testdir.makepyfile(TESTSCRIPT_OVERRIDE_FIXTURE)
    testdir.runpytest(script, "-Werror").assert_outcomes(passed=1, failed=0)


TESTSCRIPT_FIXTURE_THEN_GIVEN = """
import pytest
from hypothesis import given, strategies as st

@given(x=st.integers())
@pytest.fixture()
def test(x):
    pass
"""

pytest_version = tuple(map(int, pytest.__version__.split(".")[:2]))


def assert_outcomes(result, *, errors=0, failed=0):
    kwargs = {"errors" if pytest_version[0] > 5 else "error": errors}
    result.assert_outcomes(failed=failed, **kwargs)


def test_given_fails_if_already_decorated_with_fixture(testdir):
    script = testdir.makepyfile(TESTSCRIPT_FIXTURE_THEN_GIVEN)
    result = testdir.runpytest(script)
    if pytest_version[:2] >= (8, 4):
        assert_outcomes(result, errors=1)
    else:
        assert_outcomes(result, failed=1)


TESTSCRIPT_GIVEN_THEN_FIXTURE = """
import pytest
from hypothesis import given, strategies as st

@pytest.fixture()
@given(x=st.integers())
def test(x):
    pass
"""


def test_fixture_errors_if_already_decorated_with_given(testdir):
    script = testdir.makepyfile(TESTSCRIPT_GIVEN_THEN_FIXTURE)
    assert_outcomes(testdir.runpytest(script), errors=1)
